"""
Lskrb5Sssd - command ``ls -lan /var/lib/sss/pubconf/krb5.include.d``
====================================================================

The ``ls -lan /var/lib/sss/pubconf/krb5.include.d`` command provides information for the listing of the
``/var/lib/sss/pubconf/krb5.include.d`` directory. See the ``FileListing`` class for a more complete description
of the available features of the class.

Sample ``ls -lan /var/lib/sss/pubconf/krb5.include.d`` output::

    /var/lib/sss/pubconf/krb5.include.d:
    total 24
    drwxr-xr-x@ 6 501  20  192 Jul  1 23:46 .
    drwxr-xr-x@ 3 501  20   96 Jul  1 23:48 ..
    -rw-r--r--@ 1 501  20  674 Jul  1 23:46 domain_realm_rhidm_gwl_bz
    -rw-r--r--@ 1 501  20   35 Jul  1 23:46 krb5_libdefaults
    -rw-r--r--@ 1 501  20   98 Jul  1 23:46 localauth_plugin
    -rw-------@ 1 501  20    0 Oct  1  2021 localauth_pluginolsIe3

Examples:
    >>> '/var/lib/sss/pubconf/krb5.include.d' in ls_krb5_sssd
    True
    >>> ls_krb5_sssd.files_of('/var/lib/sss/pubconf/krb5.include.d') == ['domain_realm_rhidm_gwl_bz', 'krb5_libdefaults', 'localauth_plugin', 'localauth_pluginolsIe3']
    True
"""

from insights.specs import Specs
from .. import parser, CommandParser, FileListing
from insights.util import deprecated


@parser(Specs.ls_krb5_sssd)
class LsKrb5SSSD(CommandParser, FileListing):
    """
    .. warning::
        This class is deprecated and will be removed from 3.5.0.
        Please use the :class:`insights.parsers.ls.LSlan` instead.

    Parse the /var/lib/sss/pubconf/krb5.include.d directory listing using a standard FileListing parser.
    """
    def __init__(self, *args, **kwargs):
        deprecated(LsKrb5SSSD, "Please use the :class:`insights.parsers.ls.LSlan` instead.", "3.5.0")
        super(LsKrb5SSSD, self).__init__(*args, **kwargs)



from insights.parsers import nfs_conf as nfs_conf_parser
from insights.parsers.nfs_conf import NFSConf
from insights.tests import context_wrap
from doctest import testmod

NFS_CONF_1 = """
#
# This is a general configuration for the
# NFS daemons and tools
#
[general]
# pipefs-directory=/var/lib/nfs/rpc_pipefs
#
[exportfs]
# debug=0
#
[gssd]
# verbosity=0
# rpc-verbosity=0
# use-memcache=0
# use-machine-creds=1
use-gss-proxy=1
# avoid-dns=1
# limit-to-legacy-enctypes=0
# context-timeout=0
# rpc-timeout=5
# keytab-file=/etc/krb5.keytab
# cred-cache-directory=
# preferred-realm=
# set-home=1
# upcall-timeout=30
# cancel-timed-out-upcalls=0
#
[lockd]
# port=0
# udp-port=0
#
[mountd]
# debug=0
# manage-gids=n
# descriptors=0
# port=0
# threads=1
# reverse-lookup=n
# state-directory-path=/var/lib/nfs
# ha-callout=
# cache-use-ipaddr=n
# ttl=1800
#
[nfsdcld]
# debug=0
# storagedir=/var/lib/nfs/nfsdcld
#
[nfsdcltrack]
# debug=0
# storagedir=/var/lib/nfs/nfsdcltrack
#
[nfsd]
# debug=0
# threads=8
# host=
# port=0
# grace-time=90
# lease-time=90
# tcp=y
# vers2=n
vers3=n
# vers4=y
# vers4.0=y
# vers4.1=y
# vers4.2=y
rdma=y
rdma-port=20049
#
[statd]
# debug=0
# port=0
# outgoing-port=0
# name=
# state-directory-path=/var/lib/nfs/statd
# ha-callout=
# no-notify=0
#
[sm-notify]
# debug=0
# force=0
# retry-time=900
# outgoing-port=
# outgoing-addr=
# lift-grace=y
#
"""

NFS_CONF_2 = """
#
# This is a general configuration for the
# NFS daemons and tools
#
[general]
# pipefs-directory=/var/lib/nfs/rpc_pipefs
#
[exportfs]
# debug=0
#
[gssd]
# verbosity=0
# rpc-verbosity=0
# use-memcache=0
# use-machine-creds=1
# use-gss-proxy=1
# avoid-dns=1
# limit-to-legacy-enctypes=0
# context-timeout=0
"""

NFS_CONF_MODDOC = """
[general]
# pipefs-directory=/var/lib/nfs/rpc_pipefs

[exportfs]
debug=0

[gssd]
use-gss-proxy=1

[nfsd]
# debug=0
vers3=n
# vers4=y
# vers4.0=y
# vers4.1=y
# vers4.2=y
rdma=y
rdma-port=20049
"""


def test_nfs_conf():
    conf = NFSConf(context_wrap(NFS_CONF_1))
    assert conf

    assert conf.get('gssd', 'use-gss-proxy') == '1'
    assert conf.get('nfsd', 'rdma') == 'y'


def test_nfs_exports_empty():
    conf = NFSConf(context_wrap(NFS_CONF_2))
    assert sorted(conf.sections()) == ['exportfs', 'general', 'gssd']


def test_module_documentation():
    failed, total = testmod(nfs_conf_parser, globs={
        "nfs_conf": NFSConf(context_wrap(NFS_CONF_MODDOC))
    })
    assert failed == 0



"""
Collect all the interesting data for analysis - Core version
"""
from __future__ import absolute_import

import logging

from insights import collect
from insights.client.constants import InsightsConstants as constants
from insights.client.data_collector import DataCollector
from insights.client.utilities import systemd_notify_init_thread

APP_NAME = constants.app_name
logger = logging.getLogger(__name__)


class CoreCollector(DataCollector):
    """
    Collectoer for new core-collector
    """

    def run_collection(self, rm_conf, branch_info, blacklist_report):
        '''
        Initialize core collection here and generate the
        output directory with collected data.
        '''
        # initialize systemd-notify thread
        systemd_notify_init_thread()

        self.archive.create_archive_dir()

        logger.debug('Beginning to run core collection ...')

        collect.collect(
            tmp_path=self.archive.tmp_dir,
            archive_name=self.archive.archive_name,
            rm_conf=rm_conf or {},
            client_config=self.config,
        )

        logger.debug('Core collection finished.')

        # collect metadata
        logger.debug('Collecting metadata ...')
        self._write_branch_info(branch_info)
        self._write_display_name()
        self._write_ansible_host()
        self._write_version_info()
        self._write_tags()
        self._write_blacklist_report(blacklist_report)
        self._write_blacklisted_specs()
        self._write_egg_release()
        logger.debug('Metadata collection finished.')



"""
UdevRules - files ``/usr/lib/udev/rules.d/*`` and ``/etc/udev/rules.d/``
========================================================================

The parsers included in this module are:

UdevRulesFCWWPN - file ``/usr/lib/udev/rules.d/59-fc-wwpn-id.rules``
--------------------------------------------------------------------

UdevRules40Redhat - files ``/etc/udev/rules.d/40-redhat.rules``, ``/run/udev/rules.d/40-redhat.rules``, ``/usr/lib/udev/rules.d/40-redhat.rules``, ``/usr/local/lib/udev/rules.d/40-redhat.rules``
--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

UdevRulesOracleASM - file ``/etc/udev/rules.d/*asm*.rules``
-----------------------------------------------------------

UdevRules66MD - files ``/etc/udev/rules.d/66-md-auto-readd.rules``, ``/usr/lib/udev/rules.d/66-md-auto-readd.rules``
----------------------------------------------------------------------------------------------------------------------
"""
from insights import parser
from insights.core import LogFileOutput
from insights.specs import Specs


@parser(Specs.udev_fc_wwpn_id_rules)
class UdevRulesFCWWPN(LogFileOutput):
    """
    Read the content of ``/usr/lib/udev/rules.d/59-fc-wwpn-id.rules`` file.

    .. note::

        The syntax of the `.rules` file is complex, and no rules require to
        get the serialized parsed result currently.  An only existing rule's
        supposed to check the syntax of some specific line, so here the
        :class:`insights.core.LogFileOutput` is the base class.

    Examples:
        >>> type(udev_rules)
        <class 'insights.parsers.udev_rules.UdevRulesFCWWPN'>
        >>> 'ENV{FC_TARGET_WWPN}!="$*"; GOTO="fc_wwpn_end"' in udev_rules.lines
        True
    """
    pass


@parser(Specs.etc_udev_40_redhat_rules)
class UdevRules40Redhat(LogFileOutput):
    """
    Read the content of ``40-redhat.rules`` file.

    .. note::

        The syntax of the `.rules` file is complex, and no rules require to
        get the serialized parsed result currently.  An only existing rule's
        supposed to check the syntax of some specific line, so here the
        :class:`insights.core.LogFileOutput` is the base class.

    Sample input::

        # do not edit this file, it will be overwritten on update
        # CPU hotadd request
        SUBSYSTEM=="cpu", ACTION=="add", TEST=="online", ATTR{online}=="0", ATTR{online}="1"

        # Memory hotadd request
        SUBSYSTEM!="memory", ACTION!="add", GOTO="memory_hotplug_end"
        PROGRAM="/bin/uname -p", RESULT=="s390*", GOTO="memory_hotplug_end"

        LABEL="memory_hotplug_end"

    Examples:
        >>> 'LABEL="memory_hotplug_end"' in udev_40_redhat_rules.lines
        True
    """
    pass


@parser(Specs.etc_udev_oracle_asm_rules)
class UdevRulesOracleASM(LogFileOutput):
    """
    Read the content of ``/etc/udev/rules.d/*asm*.rules`` file.

    .. note::

        The syntax of the `.rules` file is complex, and no rules require to
        get the serialized parsed result currently.  An only existing rule's
        supposed to check the syntax of some specific lines, so here the
        :class:`insights.core.LogFileOutput` is the base class.

    Sample input::

        KERNEL=="dm*", PROGRAM=="scsi_id --page=0x83 --whitelisted --device=/dev/%k", \
        RESULT=="360060e80164c210000014c2100007a8f", \
        SYMLINK+="oracleasm/disks/asm_sbe80_7a8f", OWNER="oracle", GROUP="dba", MODE="0660"


        KERNEL=="dm*", PROGRAM=="scsi_id --page=0x83 --whitelisted --device=/dev/%k", \
        RESULT=="360060e80164c210000014c2100007a91", \
        SYMLINK+="oracleasm/disks/asm_sbe80_7a91", OWNER="oracle", GROUP="dba", MODE="0660"

        # NOTE: Insert new Oracle ASM LUN configuration before this comment
        ACTION=="add|change", KERNEL=="sd*", OPTIONS:="nowatch"

    Examples:

    >>> 'ACTION=="add|change", KERNEL=="sd*", OPTIONS:="nowatch"' in udev_oracle_asm_rules.lines
    True
    >>> udev_oracle_asm_rules.get('ACTION')[0]['raw_message']
    'ACTION=="add|change", KERNEL=="sd*", OPTIONS:="nowatch"'
    """
    pass


@parser(Specs.udev_66_md_rules)
class UdevRules66MD(LogFileOutput):
    """
    Read the content of ``66-md-auto-readd.rules`` file.

    .. note::

        The syntax of the `.rules` file is complex, and no rules require to
        get the serialized parsed result currently.  This udev rule file is
        collected with filters for some specific lines. Using
        :class:`insights.core.LogFileOutput` as the base class here.

    Sample input::

        # Enable/Disable - default is Disabled
        # to disable this rule, GOTO="md_end" should be the first active command.
        # to enable this rule, Comment out GOTO="md_end".
        GOTO="md_end"

        ACTION!="add", GOTO="md_end"
        ENV{ID_FS_TYPE}!="linux_raid_member", GOTO="md_end"
        SUBSYSTEM=="block", ACTION=="add", RUN{program}+="/sbin/md_raid_auto_readd.sh $devnode"

        #
        # Land here to exit cleanly
        LABEL="md_end"

    Examples:
        >>> 'GOTO="md_end"' in udev_66_md_rules.lines
        True
    """
    pass



"""
samba logs - files matching ``/var/log/samba/*.log``
====================================================
"""

from insights import parser, LogFileOutput
from datetime import datetime
from insights.specs import Specs


def strip_surrounds(s):
    start = 0
    end = len(s) - 1
    while (s[start] == '[' and s[end] == ']') or \
      (s[start] == '[' and s[end] == ',') or \
      (s[start] == '(' and s[end] == ')'):
        start += 1
        end -= 1
    if s[end] == ',':
        end -= 1
    return s[start:(end + 1)]


@parser(Specs.samba_logs)
class SAMBALog(LogFileOutput):
    """
    Parser class for reading samba log files.  The main work is done by the
    LogFileOutput super-class.

    .. note:

        Please refer to its super-class :class:`insights.core.LogFileOutput`

    Sample input:

       [2018/12/07 07:09:44.812154,  5, pid=6434, effective(0, 0), real(0, 0)] ../source3/param/loadparm.c:1344(free_param_opts)
        Freeing parametrics:
       [2018/12/07 07:09:44.812281,  3, pid=6434, effective(0, 0), real(0, 0)] ../source3/param/loadparm.c:547(init_globals)
        Initialising global parameters
       [2018/12/07 07:09:44.812356,  2, pid=6434, effective(0, 0), real(0, 0)] ../source3/param/loadparm.c:319(max_open_files)
        rlimit_max: increasing rlimit_max (1024) to minimum Windows limit (16384)
       [2019/45/899 11:11:04.911891,  3, pid=15822, effective(0, 0), real(0, 0)] ../source3/printing/queue_process.c:236
        (bq_sig_hup_handler) Reloading pcap cache after SIGHUP.

    Each line is parsed into a dictionary with the following keys:

        * **timestamp** - the date of the log line (as a string)
        * **datetime** - the date as a datetime object (if conversion is possible)
        * **pid** - process id of samba process being run
        * **function** - the function within the module
        * **message** - the body of the message
        * **raw_message** - the raw message before being split.

    Examples:
        >>> 'Fake' in samba_logs
        True
        >>> 'pid=15822, effective(0, 0), real(0, 0)]' in samba_logs
        True
        >>> len(samba_logs.get('Fake line')) == 1
        True
    """
    time_format = '%b %d %H:%M:%S %Y'

    def _parse_line(self, line):
        fields = line.split()
        try:
            parsed_line = {
                'timestamp': strip_surrounds(' '.join(fields[0:2])),
                'pid': strip_surrounds(fields[3]),
                'function': strip_surrounds(fields[8]),
                'message': ' '.join(fields[8:]),
                'raw_message': line
            }

            parsed_line['datetime'] = datetime.strptime(
                parsed_line['timestamp'],
                '%a %b %d %H:%M:%S %Y'
            )
        except:
            pass
        return parsed_line



"""
App manifests for use with the "--collector APP" option
Define the manifest for the app and add it to the manifests dict at the bottom
of the file
"""

malware_detection_manifest = """
# Manifest file for malware detection app data collection
---
# version is for the format of this file, not its contents.
version: 0

client:
    context:
        class: insights.core.context.HostContext
        args:
            timeout: 10 # timeout in seconds for commands. Doesn't apply to files.

    blacklist:
        files: []
        commands: []
        patterns: []
        keywords: []

    persist:
      # determines what will appear in the archive
        - name: insights.specs.default.DefaultSpecs.malware_detection
          enabled: true

    run_strategy:
        name: serial
        args:
            max_workers: null

plugins:
    # disable everything by default
    # defaults to false if not specified.
    default_component_enabled: false
    packages:
      # determines which packages are loaded. These will be namespaced to the relevant collector
        - insights.specs.datasources.malware_detection
        - insights.specs.default
    configs:
        # determines which specs get loaded
        - name: insights.specs.default.DefaultSpecs.malware_detection
          enabled: true
        # Enable specs for collecting the system's canonical facts
        - name: insights.specs.default.DefaultSpecs.mac_addresses
          enabled: true
        - name: insights.specs.Specs.mac_addresses
          enabled: true
        - name: insights.specs.default.DefaultSpecs.etc_machine_id
          enabled: true
        - name: insights.specs.Specs.etc_machine_id
          enabled: true
        - name: insights.specs.default.DefaultSpecs.hostname
          enabled: true
        - name: insights.specs.Specs.hostname
          enabled: true
        - name: insights.specs.default.DefaultSpecs.dmidecode
          enabled: true
        - name: insights.specs.Specs.dmidecode
          enabled: true
        - name: insights.specs.default.DefaultSpecs.machine_id
          enabled: true
        - name: insights.specs.Specs.machine_id
          enabled: true
        - name: insights.specs.default.DefaultSpecs.ip_addresses
          enabled: true
        - name: insights.specs.Specs.ip_addresses
          enabled: true
        - name: insights.specs.default.DefaultSpecs.subscription_manager_id
          enabled: true
        - name: insights.specs.Specs.subscription_manager_id
          enabled: true
""".lstrip()

manifests = {'malware-detection': malware_detection_manifest}
content_types = {'malware-detection': 'application/vnd.redhat.malware-detection.results+tgz'}



import doctest
import pytest

from insights.core.exceptions import SkipComponent
from insights.parsers import networkmanager_dhclient
from insights.parsers.networkmanager_dhclient import NetworkManagerDhclient
from insights.tests import context_wrap


DHCLIENT_RHEL_6 = "/etc/NetworkManager/dispatcher.d/10-dhclient"

DHCLIENT_RHEL_7 = "/etc/NetworkManager/dispatcher.d/11-dhclient"

NOT_VULNERABLE_RHEL_6 = """
#!/bin/bash
# run dhclient.d scripts in an emulated environment

PATH=/bin:/usr/bin:/sbin
SAVEDIR=/var/lib/dhclient
ETCDIR=/etc/dhcp
interface=$1

eval "$(
declare | LC_ALL=C grep '^DHCP4_[A-Z_]*=' | while read -r opt; do
    optname=${opt%%=*}
    optname=${optname,,}
    optname=new_${optname#dhcp4_}
    optvalue=${opt#*=}
    echo "$optname=$optvalue"
done
)"

[ -f /etc/sysconfig/network ] && . /etc/sysconfig/network

[ -f /etc/sysconfig/network-scripts/ifcfg-$interface ] && \
    . /etc/sysconfig/network-scripts/ifcfg-$interface

if [ -d $ETCDIR/dhclient.d ]; then
    for f in $ETCDIR/dhclient.d/*.sh; do
        if [ -x $f ]; then
            subsystem="${f%.sh}"
            subsystem="${subsystem##*/}"
            . ${f}
            if [ "$2" = "up" ]; then
                "${subsystem}_config"
            elif [ "$2" = "down" ]; then
                "${subsystem}_restore"
            fi
        fi
    done
fi
""".strip()

VULNERABLE_RHEL_7 = """
#!/bin/bash
# run dhclient.d scripts in an emulated environment

PATH=/bin:/usr/bin:/sbin
SAVEDIR=/var/lib/dhclient
ETCDIR=/etc/dhcp
interface=$1

eval "$(
declare | LC_ALL=C grep '^DHCP4_[A-Z_]*=' | while read opt; do
    optname=${opt%%=*}
    optname=${optname,,}
    optname=new_${optname#dhcp4_}
    optvalue=${opt#*=}
    echo "export $optname=$optvalue"
done
)"

[ -f /etc/sysconfig/network ] && . /etc/sysconfig/network

[ -f /etc/sysconfig/network-scripts/ifcfg-$interface ] && \
    . /etc/sysconfig/network-scripts/ifcfg-$interface

if [ -d $ETCDIR/dhclient.d ]; then
    for f in $ETCDIR/dhclient.d/*.sh; do
        if [ -x $f ]; then
            subsystem="${f%.sh}"
            subsystem="${subsystem##*/}"
            . ${f}
            if [ "$2" = "up" ]; then
                "${subsystem}_config"
            elif [ "$2" = "dhcp4-change" ]; then
                if [ "$subsystem" = "chrony" -o "$subsystem" = "ntp" ]; then
                    "${subsystem}_config"
                fi
            elif [ "$2" = "down" ]; then
                "${subsystem}_restore"
            fi
        fi
    done
fi
""".strip()


def test_no_data():
    with pytest.raises(SkipComponent):
        NetworkManagerDhclient(context_wrap(""))


def test_dhclient():
    dhclient_1 = NetworkManagerDhclient(context_wrap(VULNERABLE_RHEL_7, path=DHCLIENT_RHEL_7))
    assert dhclient_1.has_vulnerable_block

    dhclient_2 = NetworkManagerDhclient(context_wrap(NOT_VULNERABLE_RHEL_6, path=DHCLIENT_RHEL_6))
    assert not dhclient_2.has_vulnerable_block


def test_doc_examples():
    env = {
        "dhclient": NetworkManagerDhclient(context_wrap(VULNERABLE_RHEL_7, path=DHCLIENT_RHEL_7))
    }
    failed, total = doctest.testmod(networkmanager_dhclient, globs=env)
    assert failed == 0



import doctest

from insights.core.context import OSP
from insights.parsers import nova_conf
from insights.tests import context_wrap

nova_content = """

[DEFAULT]
notification_driver =
#this is comment

notification_topics=notifications

rpc_backend=rabbit
use_ipv6=True
notify_on_state_change=vm_and_task_state
notify_api_faults=False
state_path=/var/lib/nova
report_interval = 10
osapi_compute_listen=fd00:4888:1000:f901::c1
osapi_compute_workers=32
metadata_listen=fd00:4888:1000:f901::c1
metadata_workers=32
service_down_time=60
rootwrap_config=/etc/nova/rootwrap.conf
auth_strategy=keystone
use_forwarded_for=False
novncproxy_host=fd00:4888:1000:f901::c1
novncproxy_port=6080
network_api_class=nova.network.neutronv2.api.API
dhcp_domain=
security_group_api=neutron
debug=False
verbose=False
log_dir=/var/log/nova
use_syslog=False
scheduler_host_manager=nova.scheduler.host_manager.HostManager
scheduler_host_subset_size=1
cpu_allocation_ratio=16.0
disk_allocation_ratio=1.0
max_io_ops_per_host=8
max_instances_per_host=50
ram_allocation_ratio=1.0
scheduler_available_filters=nova.scheduler.filters.all_filters
scheduler_default_filters=RetryFilter,AvailabilityZoneFilter,RamFilter,ComputeFilter,ComputeCapabilitiesFilter,ImagePropertiesFilter,ServerGroupAntiAffinityFilter,ServerGroupAffinityFilter,NUMATopologyFilter,PciPassthroughFilter
scheduler_weight_classes=nova.scheduler.weights.all_weighers
scheduler_max_attempts=3
vif_plugging_is_fatal=True
vif_plugging_timeout=300
firewall_driver=nova.virt.firewall.NoopFirewallDriver
novncproxy_base_url=http://[fd00:4888:1000:f901::c1]:6080/vnc_auto.html
volume_api_class=nova.volume.cinder.API
memcached_servers=inet6:[fd00:4888:1000:f901::c0]:11211,inet6:[fd00:4888:1000:f901::c1]:11211,inet6:[fd00:4888:1000:f901::c2]:11211
[ephemeral_storage_encryption]
[glance]
api_servers=http://[fd00:4888:1000:f901::a000]:9292
[keystone_authtoken]
auth_uri=http://[fd00:4888:1000:f901::a000]:5000/v2.0
identity_uri=http://192.168.1.107:35357
admin_user=nova
admin_password=*********
admin_tenant_name=service
service_metadata_proxy=True
ovs_bridge=br-int
extension_sync_interval=600
rabbit_hosts=fd00:4888:1000:f901::c0,fd00:4888:1000:f901::c1,fd00:4888:1000:f901::c2
rabbit_use_ssl=False
rabbit_userid=guest
rabbit_password=*********
rabbit_virtual_host  =  /
rabbit_ha_queues=True
heartbeat_timeout_threshold=60
heartbeat_rate=2
"""

osp = OSP()
osp.role = "Compute"


def test_doc_examples():
    failed_count, tests = doctest.testmod(
        nova_conf,
        globs={'conf': nova_conf.NovaConf(context_wrap(nova_content))}
    )
    assert failed_count == 0


def test_nova_conf():
    result = nova_conf.NovaConf(context_wrap(nova_content, osp=osp))
    print(result)
    assert result.get("DEFAULT", "notification_driver") == ""
    assert result.get("DEFAULT", "report_interval") == "10"
    assert result.get("DEFAULT", "novncproxy_host") == "fd00:4888:1000:f901::c1"

    assert result.get("keystone_authtoken", "auth_uri") == "http://[fd00:4888:1000:f901::a000]:5000/v2.0"
    assert result.get("keystone_authtoken", "service_metadata_proxy") == "True"
    assert result.get("keystone_authtoken", "rabbit_hosts")\
        == "fd00:4888:1000:f901::c0,fd00:4888:1000:f901::c1,fd00:4888:1000:f901::c2"



""" A wrapper for the 'gpg' command::

Portions of this module are derived from A.M. Kuchling's well-designed
GPG.py, using Richard Jones' updated version 1.3, which can be found
in the pycrypto CVS repository on Sourceforge:

http://pycrypto.cvs.sourceforge.net/viewvc/pycrypto/gpg/GPG.py

This module is *not* forward-compatible with amk's; some of the
old interface has changed.  For instance, since I've added decrypt
functionality, I elected to initialize with a 'gnupghome' argument
instead of 'keyring', so that gpg can find both the public and secret
keyrings.  I've also altered some of the returned objects in order for
the caller to not have to know as much about the internals of the
result classes.

While the rest of ISconf is released under the GPL, I am releasing
this single file under the same terms that A.M. Kuchling used for
pycrypto.

Steve Traugott, stevegt@terraluna.org
Thu Jun 23 21:27:20 PDT 2005

This version of the module has been modified from Steve Traugott's version
(see http://trac.t7a.org/isconf/browser/trunk/lib/python/isconf/GPG.py) by
Vinay Sajip to make use of the subprocess module (Steve's version uses os.fork()
and so does not work on Windows). Renamed to gnupg.py to avoid confusion with
the previous versions.

Modifications Copyright (C) 2008-2019 Vinay Sajip. All rights reserved.

A unittest harness (test_gnupg.py) has also been added.
"""

__version__ = "0.4.6"
__author__ = "Vinay Sajip"
__date__  = "$17-Apr-2020 09:35:35$"

try:
    from io import StringIO
except ImportError:  # pragma: no cover
    from cStringIO import StringIO

import codecs
import locale
import logging
import os
import re
import socket
from subprocess import Popen
from subprocess import PIPE
import sys
import threading

STARTUPINFO = None
if os.name == 'nt':  # pragma: no cover
    try:
        from subprocess import STARTUPINFO, STARTF_USESHOWWINDOW, SW_HIDE
    except ImportError:
        STARTUPINFO = None

try:
    import logging.NullHandler as NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def handle(self, record):
            pass
try:
    unicode
    _py3k = False
    string_types = basestring
    text_type = unicode
except NameError:
    _py3k = True
    string_types = str
    text_type = str

logger = logging.getLogger(__name__)
if not logger.handlers:
    logger.addHandler(NullHandler())

# We use the test below because it works for Jython as well as CPython
if os.path.__name__ == 'ntpath':  # pragma: no cover
    # On Windows, we don't need shell quoting, other than worrying about
    # paths with spaces in them.
    def shell_quote(s):
        return '"%s"' % s
else:
    # Section copied from sarge

    # This regex determines which shell input needs quoting
    # because it may be unsafe
    UNSAFE = re.compile(r'[^\w%+,./:=@-]')

    def shell_quote(s):
        """
        Quote text so that it is safe for Posix command shells.

        For example, "*.py" would be converted to "'*.py'". If the text is
        considered safe it is returned unquoted.

        :param s: The value to quote
        :type s: str (or unicode on 2.x)
        :return: A safe version of the input, from the point of view of Posix
                 command shells
        :rtype: The passed-in type
        """
        if not isinstance(s, string_types):  # pragma: no cover
            raise TypeError('Expected string type, got %s' % type(s))
        if not s:
            result = "''"
        elif not UNSAFE.search(s):
            result = s
        else:
            result = "'%s'" % s.replace("'", r"'\''")
        return result

    # end of sarge code

# Now that we use shell=False, we shouldn't need to quote arguments.
# Use no_quote instead of shell_quote to remind us of where quoting
# was needed. However, note that we still need, on 2.x, to encode any
# Unicode argument with the file system encoding - see Issue #41 and
# Python issue #1759845 ("subprocess.call fails with unicode strings in
# command line").

# Allows the encoding used to be overridden in special cases by setting
# this module attribute appropriately.
fsencoding = sys.getfilesystemencoding()

def no_quote(s):
    if not _py3k and isinstance(s, text_type):
        s = s.encode(fsencoding)
    return s

def _copy_data(instream, outstream):
    # Copy one stream to another
    sent = 0
    if hasattr(sys.stdin, 'encoding'):
        enc = sys.stdin.encoding
    else:  # pragma: no cover
        enc = 'ascii'
    while True:
        # See issue #39: read can fail when e.g. a text stream is provided
        # for what is actually a binary file
        try:
            data = instream.read(1024)
        except UnicodeError:
            logger.warning('Exception occurred while reading', exc_info=1)
            break
        if not data:
            break
        sent += len(data)
        # logger.debug("sending chunk (%d): %r", sent, data[:256])
        try:
            outstream.write(data)
        except UnicodeError:  # pragma: no cover
            outstream.write(data.encode(enc))
        except:
            # Can sometimes get 'broken pipe' errors even when the data has all
            # been sent
            logger.exception('Error sending data')
            break
    try:
        outstream.close()
    except IOError:  # pragma: no cover
        logger.warning('Exception occurred while closing: ignored', exc_info=1)
    logger.debug("closed output, %d bytes sent", sent)

def _threaded_copy_data(instream, outstream):
    wr = threading.Thread(target=_copy_data, args=(instream, outstream))
    wr.setDaemon(True)
    logger.debug('data copier: %r, %r, %r', wr, instream, outstream)
    wr.start()
    return wr

def _write_passphrase(stream, passphrase, encoding):
    passphrase = '%s\n' % passphrase
    passphrase = passphrase.encode(encoding)
    stream.write(passphrase)
    logger.debug('Wrote passphrase')

def _is_sequence(instance):
    return isinstance(instance, (list, tuple, set, frozenset))

def _make_memory_stream(s):
    try:
        from io import BytesIO
        rv = BytesIO(s)
    except ImportError:  # pragma: no cover
        rv = StringIO(s)
    return rv

def _make_binary_stream(s, encoding):
    if _py3k:
        if isinstance(s, str):
            s = s.encode(encoding)
    else:
        if type(s) is not str:
            s = s.encode(encoding)
    return _make_memory_stream(s)

class Verify(object):
    "Handle status messages for --verify"

    TRUST_UNDEFINED = 0
    TRUST_NEVER = 1
    TRUST_MARGINAL = 2
    TRUST_FULLY = 3
    TRUST_ULTIMATE = 4

    TRUST_LEVELS = {
        "TRUST_UNDEFINED" : TRUST_UNDEFINED,
        "TRUST_NEVER" : TRUST_NEVER,
        "TRUST_MARGINAL" : TRUST_MARGINAL,
        "TRUST_FULLY" : TRUST_FULLY,
        "TRUST_ULTIMATE" : TRUST_ULTIMATE,
    }

    # for now, just the most common error codes. This can be expanded as and
    # when reports come in of other errors.
    GPG_SYSTEM_ERROR_CODES = {
        1: 'permission denied',
        35: 'file exists',
        81: 'file not found',
        97: 'not a directory',
    }

    GPG_ERROR_CODES = {
        11: 'incorrect passphrase',
    }

    def __init__(self, gpg):
        self.gpg = gpg
        self.valid = False
        self.fingerprint = self.creation_date = self.timestamp = None
        self.signature_id = self.key_id = None
        self.username = None
        self.key_id = None
        self.key_status = None
        self.status = None
        self.pubkey_fingerprint = None
        self.expire_timestamp = None
        self.sig_timestamp = None
        self.trust_text = None
        self.trust_level = None
        self.sig_info = {}

    def __nonzero__(self):
        return self.valid

    __bool__ = __nonzero__

    def handle_status(self, key, value):

        def update_sig_info(**kwargs):
            sig_id = self.signature_id
            if sig_id:
                info = self.sig_info[sig_id]
                info.update(kwargs)

        if key in self.TRUST_LEVELS:
            self.trust_text = key
            self.trust_level = self.TRUST_LEVELS[key]
            update_sig_info(trust_level=self.trust_level,
                            trust_text=self.trust_text)
        elif key in ("WARNING", "ERROR"):
            logger.warning('potential problem: %s: %s', key, value)
        elif key == "BADSIG":  # pragma: no cover
            self.valid = False
            self.status = 'signature bad'
            self.key_id, self.username = value.split(None, 1)
            update_sig_info(keyid=self.key_id, username=self.username,
                            status=self.status)
        elif key == "ERRSIG":  # pragma: no cover
            self.valid = False
            parts = value.split()
            (self.key_id,
             algo, hash_algo,
             cls,
             self.timestamp) = parts[:5]
            # Since GnuPG 2.2.7, a fingerprint is tacked on
            if len(parts) >= 7:
                self.fingerprint = parts[6]
            self.status = 'signature error'
            update_sig_info(keyid=self.key_id, timestamp=self.timestamp,
                            fingerprint=self.fingerprint, status=self.status)
        elif key == "EXPSIG":  # pragma: no cover
            self.valid = False
            self.status = 'signature expired'
            self.key_id, self.username = value.split(None, 1)
            update_sig_info(keyid=self.key_id, username=self.username,
                            status=self.status)
        elif key == "GOODSIG":
            self.valid = True
            self.status = 'signature good'
            self.key_id, self.username = value.split(None, 1)
            update_sig_info(keyid=self.key_id, username=self.username,
                            status=self.status)
        elif key == "VALIDSIG":
            fingerprint, creation_date, sig_ts, expire_ts = value.split()[:4]
            (self.fingerprint,
             self.creation_date,
             self.sig_timestamp,
             self.expire_timestamp) = (fingerprint, creation_date, sig_ts,
                                       expire_ts)
            # may be different if signature is made with a subkey
            self.pubkey_fingerprint = value.split()[-1]
            self.status = 'signature valid'
            update_sig_info(fingerprint=fingerprint, creation_date=creation_date,
                            timestamp=sig_ts, expiry=expire_ts,
                            pubkey_fingerprint=self.pubkey_fingerprint,
                            status=self.status)
        elif key == "SIG_ID":
            sig_id, creation_date, timestamp = value.split()
            self.sig_info[sig_id] = {'creation_date': creation_date,
                                     'timestamp': timestamp}
            (self.signature_id,
             self.creation_date, self.timestamp) = (sig_id, creation_date,
                                                    timestamp)
        elif key == "DECRYPTION_FAILED":  # pragma: no cover
            self.valid = False
            self.key_id = value
            self.status = 'decryption failed'
        elif key == "NO_PUBKEY":  # pragma: no cover
            self.valid = False
            self.key_id = value
            self.status = 'no public key'
        elif key in ("EXPKEYSIG", "REVKEYSIG"):  # pragma: no cover
            # signed with expired or revoked key
            self.valid = False
            self.key_id = value.split()[0]
            if key == "EXPKEYSIG":
                self.key_status = 'signing key has expired'
            else:
                self.key_status = 'signing key was revoked'
            self.status = self.key_status
            update_sig_info(status=self.status, keyid=self.key_id)
        elif key in ("UNEXPECTED", "FAILURE"):  # pragma: no cover
            self.valid = False
            self.key_id = value
            if key == "UNEXPECTED":
                self.status = 'unexpected data'
            else:
                # N.B. there might be other reasons. For example, if an output
                # file can't  be created - /dev/null/foo will lead to a
                # "not a directory" error, but which is not sent as a status
                # message with the [GNUPG:] prefix. Similarly if you try to
                # write to "/etc/foo" as a non-root user, a "permission denied"
                # error will be sent as a non-status message.
                message = 'error - %s' % value
                operation, code = value.rsplit(' ', 1)
                if code.isdigit():
                    code = int(code) & 0xFFFFFF  # lose the error source
                    if self.gpg.error_map and code in self.gpg.error_map:
                        message = '%s: %s' % (operation, self.gpg.error_map[code])
                    else:
                        system_error = bool(code & 0x8000)
                        code = code & 0x7FFF
                        if system_error:
                            mapping = self.GPG_SYSTEM_ERROR_CODES
                        else:
                            mapping = self.GPG_ERROR_CODES
                        if code in mapping:
                            message = '%s: %s' % (operation, mapping[code])
                if not self.status:
                    self.status = message
        elif key in ("DECRYPTION_INFO", "PLAINTEXT", "PLAINTEXT_LENGTH",
                     "NO_SECKEY", "BEGIN_SIGNING"):
            pass
        else:  # pragma: no cover
            logger.debug('message ignored: %s, %s', key, value)

class ImportResult(object):
    "Handle status messages for --import"

    counts = '''count no_user_id imported imported_rsa unchanged
            n_uids n_subk n_sigs n_revoc sec_read sec_imported
            sec_dups not_imported'''.split()
    def __init__(self, gpg):
        self.gpg = gpg
        self.results = []
        self.fingerprints = []
        for result in self.counts:
            setattr(self, result, 0)

    def __nonzero__(self):
        if self.not_imported: return False
        if not self.fingerprints: return False
        return True

    __bool__ = __nonzero__

    ok_reason = {
        '0': 'Not actually changed',
        '1': 'Entirely new key',
        '2': 'New user IDs',
        '4': 'New signatures',
        '8': 'New subkeys',
        '16': 'Contains private key',
    }

    problem_reason = {
        '0': 'No specific reason given',
        '1': 'Invalid Certificate',
        '2': 'Issuer Certificate missing',
        '3': 'Certificate Chain too long',
        '4': 'Error storing certificate',
    }

    def handle_status(self, key, value):
        if key in ("WARNING", "ERROR"):
            logger.warning('potential problem: %s: %s', key, value)
        elif key in ("IMPORTED", "KEY_CONSIDERED"):
            # this duplicates info we already see in import_ok & import_problem
            pass
        elif key == "NODATA":  # pragma: no cover
            self.results.append({'fingerprint': None,
                'problem': '0', 'text': 'No valid data found'})
        elif key == "IMPORT_OK":
            reason, fingerprint = value.split()
            reasons = []
            for code, text in list(self.ok_reason.items()):
                if int(reason) | int(code) == int(reason):
                    reasons.append(text)
            reasontext = '\n'.join(reasons) + "\n"
            self.results.append({'fingerprint': fingerprint,
                'ok': reason, 'text': reasontext})
            self.fingerprints.append(fingerprint)
        elif key == "IMPORT_PROBLEM":  # pragma: no cover
            try:
                reason, fingerprint = value.split()
            except:
                reason = value
                fingerprint = '<unknown>'
            self.results.append({'fingerprint': fingerprint,
                'problem': reason, 'text': self.problem_reason[reason]})
        elif key == "IMPORT_RES":
            import_res = value.split()
            for i, count in enumerate(self.counts):
                setattr(self, count, int(import_res[i]))
        elif key == "KEYEXPIRED":  # pragma: no cover
            self.results.append({'fingerprint': None,
                'problem': '0', 'text': 'Key expired'})
        elif key == "SIGEXPIRED":  # pragma: no cover
            self.results.append({'fingerprint': None,
                'problem': '0', 'text': 'Signature expired'})
        elif key == "FAILURE":  # pragma: no cover
            self.results.append({'fingerprint': None,
                'problem': '0', 'text': 'Other failure'})
        else:  # pragma: no cover
            logger.debug('message ignored: %s, %s', key, value)

    def summary(self):
        result = []
        result.append('%d imported' % self.imported)
        if self.not_imported:  # pragma: no cover
            result.append('%d not imported' % self.not_imported)
        return ', '.join(result)

ESCAPE_PATTERN = re.compile(r'\\x([0-9a-f][0-9a-f])', re.I)
BASIC_ESCAPES = {
    r'\n': '\n',
    r'\r': '\r',
    r'\f': '\f',
    r'\v': '\v',
    r'\b': '\b',
    r'\0': '\0',
}

class SendResult(object):
    def __init__(self, gpg):
        self.gpg = gpg

    def handle_status(self, key, value):
        logger.debug('SendResult: %s: %s', key, value)

def _set_fields(target, fieldnames, args):
    for i, var in enumerate(fieldnames):
        if i < len(args):
            target[var] = args[i]
        else:
            target[var] = 'unavailable'

class SearchKeys(list):
    ''' Handle status messages for --search-keys.

        Handle pub and uid (relating the latter to the former).

        Don't care about the rest
    '''

    UID_INDEX = 1
    FIELDS = 'type keyid algo length date expires'.split()

    def __init__(self, gpg):
        self.gpg = gpg
        self.curkey = None
        self.fingerprints = []
        self.uids = []

    def get_fields(self, args):
        result = {}
        _set_fields(result, self.FIELDS, args)
        result['uids'] = []
        result['sigs'] = []
        return result

    def pub(self, args):
        self.curkey = curkey = self.get_fields(args)
        self.append(curkey)

    def uid(self, args):
        uid = args[self.UID_INDEX]
        uid = ESCAPE_PATTERN.sub(lambda m: chr(int(m.group(1), 16)), uid)
        for k, v in BASIC_ESCAPES.items():
            uid = uid.replace(k, v)
        self.curkey['uids'].append(uid)
        self.uids.append(uid)

    def handle_status(self, key, value):  # pragma: no cover
        pass

class ListKeys(SearchKeys):
    ''' Handle status messages for --list-keys, --list-sigs.

        Handle pub and uid (relating the latter to the former).

        Don't care about (info from src/DETAILS):

        crt = X.509 certificate
        crs = X.509 certificate and private key available
        uat = user attribute (same as user id except for field 10).
        sig = signature
        rev = revocation signature
        pkd = public key data (special field format, see below)
        grp = reserved for gpgsm
        rvk = revocation key
    '''

    UID_INDEX = 9
    FIELDS = 'type trust length algo keyid date expires dummy ownertrust uid sig cap issuer flag token hash curve compliance updated origin'.split()

    def __init__(self, gpg):
        super(ListKeys, self).__init__(gpg)
        self.in_subkey = False
        self.key_map = {}

    def key(self, args):
        self.curkey = curkey = self.get_fields(args)
        if curkey['uid']:
            curkey['uids'].append(curkey['uid'])
        del curkey['uid']
        curkey['subkeys'] = []
        self.append(curkey)
        self.in_subkey = False

    pub = sec = key

    def fpr(self, args):
        fp = args[9]
        if fp in self.key_map and self.gpg.check_fingerprint_collisions:  # pragma: no cover
            raise ValueError('Unexpected fingerprint collision: %s' % fp)
        if not self.in_subkey:
            self.curkey['fingerprint'] = fp
            self.fingerprints.append(fp)
            self.key_map[fp] = self.curkey
        else:
            self.curkey['subkeys'][-1].append(fp)
            self.key_map[fp] = self.curkey

    def _collect_subkey_info(self, curkey, args):
        info_map = curkey.setdefault('subkey_info', {})
        info = {}
        _set_fields(info, self.FIELDS, args)
        info_map[args[4]] = info

    def sub(self, args):
        # See issue #81. We create a dict with more information about
        # subkeys, but for backward compatibility reason, have to add it in
        # as a separate entry 'subkey_info'
        subkey = [args[4], args[11]]    # keyid, type
        self.curkey['subkeys'].append(subkey)
        self._collect_subkey_info(self.curkey, args)
        self.in_subkey = True

    def ssb(self, args):
        subkey = [args[4], None]    # keyid, type
        self.curkey['subkeys'].append(subkey)
        self._collect_subkey_info(self.curkey, args)
        self.in_subkey = True

    def sig(self, args):
        # keyid, uid, sigclass
        self.curkey['sigs'].append((args[4], args[9], args[10]))

class ScanKeys(ListKeys):
    ''' Handle status messages for --with-fingerprint.'''

    def sub(self, args):
        # --with-fingerprint --with-colons somehow outputs fewer colons,
        # use the last value args[-1] instead of args[11]
        subkey = [args[4], args[-1]]
        self.curkey['subkeys'].append(subkey)
        self._collect_subkey_info(self.curkey, args)
        self.in_subkey = True

class TextHandler(object):
    def _as_text(self):
        return self.data.decode(self.gpg.encoding, self.gpg.decode_errors)

    if _py3k:
        __str__ = _as_text
    else:
        __unicode__ = _as_text

        def __str__(self):
            return self.data


class Crypt(Verify, TextHandler):
    "Handle status messages for --encrypt and --decrypt"
    def __init__(self, gpg):
        Verify.__init__(self, gpg)
        self.data = ''
        self.ok = False
        self.status = ''
        self.key_id = None

    def __nonzero__(self):
        if self.ok: return True
        return False

    __bool__ = __nonzero__

    def handle_status(self, key, value):
        if key in ("WARNING", "ERROR"):
            logger.warning('potential problem: %s: %s', key, value)
        elif key == "NODATA":
            self.status = "no data was provided"
        elif key in ("NEED_PASSPHRASE", "BAD_PASSPHRASE", "GOOD_PASSPHRASE",
                     "MISSING_PASSPHRASE", "DECRYPTION_FAILED",
                     "KEY_NOT_CREATED", "NEED_PASSPHRASE_PIN"):
            self.status = key.replace("_", " ").lower()
        elif key == "NEED_PASSPHRASE_SYM":
            self.status = 'need symmetric passphrase'
        elif key == "BEGIN_DECRYPTION":
            self.status = 'decryption incomplete'
        elif key == "BEGIN_ENCRYPTION":
            self.status = 'encryption incomplete'
        elif key == "DECRYPTION_OKAY":
            self.status = 'decryption ok'
            self.ok = True
        elif key == "END_ENCRYPTION":
            self.status = 'encryption ok'
            self.ok = True
        elif key == "INV_RECP":  # pragma: no cover
            self.status = 'invalid recipient'
        elif key == "KEYEXPIRED":  # pragma: no cover
            self.status = 'key expired'
        elif key == "SIG_CREATED":  # pragma: no cover
            self.status = 'sig created'
        elif key == "SIGEXPIRED":  # pragma: no cover
            self.status = 'sig expired'
        elif key == "ENC_TO":  # pragma: no cover
            # ENC_TO <long_keyid> <keytype> <keylength>
            self.key_id = value.split(' ', 1)[0]
        elif key in ("USERID_HINT", "GOODMDC",
                     "END_DECRYPTION", "CARDCTRL", "BADMDC",
                     "SC_OP_FAILURE", "SC_OP_SUCCESS",
                     "PINENTRY_LAUNCHED", "KEY_CONSIDERED"):
            pass
        else:
            Verify.handle_status(self, key, value)

class GenKey(object):
    "Handle status messages for --gen-key"
    def __init__(self, gpg):
        self.gpg = gpg
        self.type = None
        self.fingerprint = None

    def __nonzero__(self):
        if self.fingerprint: return True
        return False

    __bool__ = __nonzero__

    def __str__(self):
        return self.fingerprint or ''

    def handle_status(self, key, value):
        if key in ("WARNING", "ERROR"):  # pragma: no cover
            logger.warning('potential problem: %s: %s', key, value)
        elif key == "KEY_CREATED":
            (self.type,self.fingerprint) = value.split()
        elif key in ("PROGRESS", "GOOD_PASSPHRASE", "KEY_NOT_CREATED"):
            pass
        else:  # pragma: no cover
            logger.debug('message ignored: %s, %s', key, value)

class ExportResult(GenKey):
    """Handle status messages for --export[-secret-key].

    For now, just use an existing class to base it on - if needed, we
    can override handle_status for more specific message handling.
    """
    def handle_status(self, key, value):
        if key in ("EXPORTED", "EXPORT_RES"):
            pass
        else:
            super(ExportResult, self).handle_status(key, value)

class DeleteResult(object):
    "Handle status messages for --delete-key and --delete-secret-key"
    def __init__(self, gpg):
        self.gpg = gpg
        self.status = 'ok'

    def __str__(self):
        return self.status

    problem_reason = {
        '1': 'No such key',
        '2': 'Must delete secret key first',
        '3': 'Ambiguous specification',
    }

    def handle_status(self, key, value):
        if key == "DELETE_PROBLEM":  # pragma: no cover
            self.status = self.problem_reason.get(value,
                                                  "Unknown error: %r" % value)
        else:  # pragma: no cover
            logger.debug('message ignored: %s, %s', key, value)

    def __nonzero__(self):
        return self.status == 'ok'

    __bool__ = __nonzero__


class TrustResult(DeleteResult):
    pass


class Sign(TextHandler):
    "Handle status messages for --sign"
    def __init__(self, gpg):
        self.gpg = gpg
        self.type = None
        self.hash_algo = None
        self.fingerprint = None
        self.status = None
        self.key_id = None
        self.username = None

    def __nonzero__(self):
        return self.fingerprint is not None

    __bool__ = __nonzero__

    def handle_status(self, key, value):
        if key in ("WARNING", "ERROR", "FAILURE"):  # pragma: no cover
            logger.warning('potential problem: %s: %s', key, value)
        elif key in ("KEYEXPIRED", "SIGEXPIRED"):  # pragma: no cover
            self.status = 'key expired'
        elif key == "KEYREVOKED":  # pragma: no cover
            self.status = 'key revoked'
        elif key == "SIG_CREATED":
            (self.type,
             algo, self.hash_algo, cls, self.timestamp, self.fingerprint
            ) = value.split()
            self.status = 'signature created'
        elif key == "USERID_HINT":  # pragma: no cover
            self.key_id, self.username = value.split(' ', 1)
        elif key == "BAD_PASSPHRASE":
            self.status = 'bad passphrase'
        elif key in ("NEED_PASSPHRASE", "GOOD_PASSPHRASE", "BEGIN_SIGNING"):
            pass
        else:  # pragma: no cover
            logger.debug('message ignored: %s, %s', key, value)

VERSION_RE = re.compile(r'gpg \(GnuPG(?:/MacGPG2)?\) (\d+(\.\d+)*)'.encode('ascii'), re.I)
HEX_DIGITS_RE = re.compile(r'[0-9a-f]+$', re.I)

class GPG(object):

    error_map = None

    decode_errors = 'strict'

    result_map = {
        'crypt': Crypt,
        'delete': DeleteResult,
        'generate': GenKey,
        'import': ImportResult,
        'send': SendResult,
        'list': ListKeys,
        'scan': ScanKeys,
        'search': SearchKeys,
        'sign': Sign,
        'trust': TrustResult,
        'verify': Verify,
        'export': ExportResult,
    }

    "Encapsulate access to the gpg executable"
    def __init__(self, gpgbinary='gpg', gnupghome=None, verbose=False,
                 use_agent=False, keyring=None, options=None,
                 secret_keyring=None):
        """Initialize a GPG process wrapper.  Options are:

        gpgbinary -- full pathname for GPG binary.

        gnupghome -- full pathname to where we can find the public and
        private keyrings.  Default is whatever gpg defaults to.
        keyring -- name of alternative keyring file to use, or list of such
        keyrings. If specified, the default keyring is not used.
        options =-- a list of additional options to pass to the GPG binary.
        secret_keyring -- name of alternative secret keyring file to use, or
        list of such keyrings.
        """
        self.gpgbinary = gpgbinary
        self.gnupghome = gnupghome
        # issue 112: fail if the specified value isn't a directory
        if gnupghome and not os.path.isdir(gnupghome):
            raise ValueError('gnupghome should be a directory (it isn\'t): %s' % gnupghome)
        if keyring:
            # Allow passing a string or another iterable. Make it uniformly
            # a list of keyring filenames
            if isinstance(keyring, string_types):
                keyring = [keyring]
        self.keyring = keyring
        if secret_keyring:
            # Allow passing a string or another iterable. Make it uniformly
            # a list of keyring filenames
            if isinstance(secret_keyring, string_types):
                secret_keyring = [secret_keyring]
        self.secret_keyring = secret_keyring
        self.verbose = verbose
        self.use_agent = use_agent
        if isinstance(options, str):  # pragma: no cover
            options = [options]
        self.options = options
        self.on_data = None  # or a callable - will be called with data chunks
        # Changed in 0.3.7 to use Latin-1 encoding rather than
        # locale.getpreferredencoding falling back to sys.stdin.encoding
        # falling back to utf-8, because gpg itself uses latin-1 as the default
        # encoding.
        self.encoding = 'latin-1'
        if gnupghome and not os.path.isdir(self.gnupghome):
            os.makedirs(self.gnupghome,0x1C0)
        try:
            p = self._open_subprocess(["--version"])
        except OSError:
            msg = 'Unable to run gpg (%s) - it may not be available.' % self.gpgbinary
            logger.exception(msg)
            raise OSError(msg)
        result = self.result_map['verify'](self) # any result will do for this
        self._collect_output(p, result, stdin=p.stdin)
        if p.returncode != 0:  # pragma: no cover
            raise ValueError("Error invoking gpg: %s: %s" % (p.returncode,
                                                             result.stderr))
        m = VERSION_RE.match(result.data)
        if not m:  # pragma: no cover
            self.version = None
        else:
            dot = '.'.encode('ascii')
            self.version = tuple([int(s) for s in m.groups()[0].split(dot)])

        # See issue #97. It seems gpg allow duplicate keys in keyrings, so we
        # can't be too strict.
        self.check_fingerprint_collisions = False

    def make_args(self, args, passphrase):
        """
        Make a list of command line elements for GPG. The value of ``args``
        will be appended. The ``passphrase`` argument needs to be True if
        a passphrase will be sent to GPG, else False.
        """
        cmd = [self.gpgbinary, '--status-fd', '2', '--no-tty', '--no-verbose']
        if 'DEBUG_IPC' in os.environ:
            cmd.extend(['--debug', 'ipc'])
        if passphrase and hasattr(self, 'version'):
            if self.version >= (2, 1):
                cmd[1:1] = ['--pinentry-mode', 'loopback']
        cmd.extend(['--fixed-list-mode', '--batch', '--with-colons'])
        if self.gnupghome:
            cmd.extend(['--homedir',  no_quote(self.gnupghome)])
        if self.keyring:
            cmd.append('--no-default-keyring')
            for fn in self.keyring:
                cmd.extend(['--keyring', no_quote(fn)])
        if self.secret_keyring:
            for fn in self.secret_keyring:
                cmd.extend(['--secret-keyring', no_quote(fn)])
        if passphrase:
            cmd.extend(['--passphrase-fd', '0'])
        if self.use_agent:  # pragma: no cover
            cmd.append('--use-agent')
        if self.options:
            cmd.extend(self.options)
        cmd.extend(args)
        return cmd

    def _open_subprocess(self, args, passphrase=False):
        # Internal method: open a pipe to a GPG subprocess and return
        # the file objects for communicating with it.

        # def debug_print(cmd):
            # result = []
            # for c in cmd:
                # if ' ' not in c:
                    # result.append(c)
                # else:
                    # if '"' not in c:
                        # result.append('"%s"' % c)
                    # elif "'" not in c:
                        # result.append("'%s'" % c)
                    # else:
                        # result.append(c)  # give up
            # return ' '.join(cmd)
        from subprocess import list2cmdline as debug_print

        cmd = self.make_args(args, passphrase)
        if self.verbose:  # pragma: no cover
            print(debug_print(cmd))
        if not STARTUPINFO:
            si = None
        else:  # pragma: no cover
            si = STARTUPINFO()
            si.dwFlags = STARTF_USESHOWWINDOW
            si.wShowWindow = SW_HIDE
        result = Popen(cmd, shell=False, stdin=PIPE, stdout=PIPE, stderr=PIPE,
                       startupinfo=si)
        logger.debug("%s: %s", result.pid, debug_print(cmd))
        return result

    def _read_response(self, stream, result):
        # Internal method: reads all the stderr output from GPG, taking notice
        # only of lines that begin with the magic [GNUPG:] prefix.
        #
        # Calls methods on the response object for each valid token found,
        # with the arg being the remainder of the status line.
        lines = []
        while True:
            line = stream.readline()
            if len(line) == 0:
                break
            lines.append(line)
            line = line.rstrip()
            if self.verbose:  # pragma: no cover
                print(line)
            logger.debug("%s", line)
            if line[0:9] == '[GNUPG:] ':
                # Chop off the prefix
                line = line[9:]
                L = line.split(None, 1)
                keyword = L[0]
                if len(L) > 1:
                    value = L[1]
                else:
                    value = ""
                result.handle_status(keyword, value)
        result.stderr = ''.join(lines)

    def _read_data(self, stream, result, on_data=None):
        # Read the contents of the file from GPG's stdout
        chunks = []
        while True:
            data = stream.read(1024)
            if len(data) == 0:
                if on_data:
                    on_data(data)
                break
            logger.debug("chunk: %r" % data[:256])
            append = True
            if on_data:
                append = on_data(data) != False
            if append:
                chunks.append(data)
        if _py3k:
            # Join using b'' or '', as appropriate
            result.data = type(data)().join(chunks)
        else:
            result.data = ''.join(chunks)

    def _collect_output(self, process, result, writer=None, stdin=None):
        """
        Drain the subprocesses output streams, writing the collected output
        to the result. If a writer thread (writing to the subprocess) is given,
        make sure it's joined before returning. If a stdin stream is given,
        close it before returning.
        """
        stderr = codecs.getreader(self.encoding)(process.stderr)
        rr = threading.Thread(target=self._read_response, args=(stderr, result))
        rr.setDaemon(True)
        logger.debug('stderr reader: %r', rr)
        rr.start()

        stdout = process.stdout
        dr = threading.Thread(target=self._read_data, args=(stdout, result,
                              self.on_data))
        dr.setDaemon(True)
        logger.debug('stdout reader: %r', dr)
        dr.start()

        dr.join()
        rr.join()
        if writer is not None:
            writer.join()
        process.wait()
        rc = process.returncode
        if rc != 0:
            logger.warning('gpg returned a non-zero error code: %d', rc)
        if stdin is not None:
            try:
                stdin.close()
            except IOError:  # pragma: no cover
                pass
        stderr.close()
        stdout.close()

    def _handle_io(self, args, fileobj, result, passphrase=None, binary=False):
        "Handle a call to GPG - pass input data, collect output data"
        # Handle a basic data call - pass data to GPG, handle the output
        # including status information. Garbage In, Garbage Out :)
        p = self._open_subprocess(args, passphrase is not None)
        if not binary:  # pragma: no cover
            stdin = codecs.getwriter(self.encoding)(p.stdin)
        else:
            stdin = p.stdin
        if passphrase:
            _write_passphrase(stdin, passphrase, self.encoding)
        writer = _threaded_copy_data(fileobj, stdin)
        self._collect_output(p, result, writer, stdin)
        return result

    #
    # SIGNATURE METHODS
    #
    def sign(self, message, **kwargs):
        """sign message"""
        f = _make_binary_stream(message, self.encoding)
        result = self.sign_file(f, **kwargs)
        f.close()
        return result

    def set_output_without_confirmation(self, args, output):
        "If writing to a file which exists, avoid a confirmation message."
        if os.path.exists(output):
            # We need to avoid an overwrite confirmation message
            args.extend(['--yes'])
        args.extend(['--output', no_quote(output)])

    def is_valid_passphrase(self, passphrase):
        """
        Confirm that the passphrase doesn't contain newline-type characters -
        it is passed in a pipe to gpg, and so not checking could lead to
        spoofing attacks by passing arbitrary text after passphrase and newline.
        """
        return ('\n' not in passphrase and '\r' not in passphrase and
                '\x00' not in passphrase)

    def sign_file(self, file, keyid=None, passphrase=None, clearsign=True,
                  detach=False, binary=False, output=None, extra_args=None):
        """sign file"""
        if passphrase and not self.is_valid_passphrase(passphrase):
            raise ValueError('Invalid passphrase')
        logger.debug("sign_file: %s", file)
        if binary:  # pragma: no cover
            args = ['-s']
        else:
            args = ['-sa']
        # You can't specify detach-sign and clearsign together: gpg ignores
        # the detach-sign in that case.
        if detach:
            args.append("--detach-sign")
        elif clearsign:
            args.append("--clearsign")
        if keyid:
            args.extend(['--default-key', no_quote(keyid)])
        if output:  # write the output to a file with the specified name
            self.set_output_without_confirmation(args, output)

        if extra_args:
            args.extend(extra_args)
        result = self.result_map['sign'](self)
        #We could use _handle_io here except for the fact that if the
        #passphrase is bad, gpg bails and you can't write the message.
        p = self._open_subprocess(args, passphrase is not None)
        try:
            stdin = p.stdin
            if passphrase:
                _write_passphrase(stdin, passphrase, self.encoding)
            writer = _threaded_copy_data(file, stdin)
        except IOError:  # pragma: no cover
            logging.exception("error writing message")
            writer = None
        self._collect_output(p, result, writer, stdin)
        return result

    def verify(self, data, **kwargs):
        """Verify the signature on the contents of the string 'data'

        >>> GPGBINARY = os.environ.get('GPGBINARY', 'gpg')
        >>> if not os.path.isdir('keys'): os.mkdir('keys')
        >>> gpg = GPG(gpgbinary=GPGBINARY, gnupghome='keys')
        >>> input = gpg.gen_key_input(passphrase='foo')
        >>> key = gpg.gen_key(input)
        >>> assert key
        >>> sig = gpg.sign('hello',keyid=key.fingerprint,passphrase='bar')
        >>> assert not sig
        >>> sig = gpg.sign('hello',keyid=key.fingerprint,passphrase='foo')
        >>> assert sig
        >>> verify = gpg.verify(sig.data)
        >>> assert verify

        """
        f = _make_binary_stream(data, self.encoding)
        result = self.verify_file(f, **kwargs)
        f.close()
        return result

    def verify_file(self, file, data_filename=None, close_file=True, extra_args=None):
        "Verify the signature on the contents of the file-like object 'file'"
        logger.debug('verify_file: %r, %r', file, data_filename)
        result = self.result_map['verify'](self)
        args = ['--verify']
        if extra_args:
            args.extend(extra_args)
        if data_filename is None:
            self._handle_io(args, file, result, binary=True)
        else:
            logger.debug('Handling detached verification')
            import tempfile
            fd, fn = tempfile.mkstemp(prefix='pygpg')
            s = file.read()
            if close_file:
                file.close()
            logger.debug('Wrote to temp file: %r', s)
            os.write(fd, s)
            os.close(fd)
            args.append(no_quote(fn))
            args.append(no_quote(data_filename))
            try:
                p = self._open_subprocess(args)
                self._collect_output(p, result, stdin=p.stdin)
            finally:
                os.unlink(fn)
        return result

    def verify_data(self, sig_filename, data, extra_args=None):
        "Verify the signature in sig_filename against data in memory"
        logger.debug('verify_data: %r, %r ...', sig_filename, data[:16])
        result = self.result_map['verify'](self)
        args = ['--verify']
        if extra_args:
            args.extend(extra_args)
        args.extend([no_quote(sig_filename), '-'])
        stream = _make_memory_stream(data)
        self._handle_io(args, stream, result, binary=True)
        return result

    #
    # KEY MANAGEMENT
    #

    def import_keys(self, key_data, extra_args=None):
        """
        Import the key_data into our keyring.
        """
        result = self.result_map['import'](self)
        logger.debug('import_keys: %r', key_data[:256])
        data = _make_binary_stream(key_data, self.encoding)
        args = ['--import']
        if extra_args:
            args.extend(extra_args)
        self._handle_io(args, data, result, binary=True)
        logger.debug('import_keys result: %r', result.__dict__)
        data.close()
        return result

    def recv_keys(self, keyserver, *keyids):
        """Import a key from a keyserver

        >>> import shutil
        >>> shutil.rmtree("keys", ignore_errors=True)
        >>> GPGBINARY = os.environ.get('GPGBINARY', 'gpg')
        >>> if not os.path.isdir('keys'): os.mkdir('keys')
        >>> gpg = GPG(gpgbinary=GPGBINARY, gnupghome='keys')
        >>> os.chmod('keys', 0x1C0)
        >>> result = gpg.recv_keys('pgp.mit.edu', '92905378')
        >>> if 'NO_EXTERNAL_TESTS' not in os.environ: assert result

        """
        result = self.result_map['import'](self)
        logger.debug('recv_keys: %r', keyids)
        data = _make_binary_stream("", self.encoding)
        #data = ""
        args = ['--keyserver', no_quote(keyserver), '--recv-keys']
        args.extend([no_quote(k) for k in keyids])
        self._handle_io(args, data, result, binary=True)
        logger.debug('recv_keys result: %r', result.__dict__)
        data.close()
        return result

    def send_keys(self, keyserver, *keyids):
        """Send a key to a keyserver.

        Note: it's not practical to test this function without sending
        arbitrary data to live keyservers.
        """
        result = self.result_map['send'](self)
        logger.debug('send_keys: %r', keyids)
        data = _make_binary_stream('', self.encoding)
        #data = ""
        args = ['--keyserver', no_quote(keyserver), '--send-keys']
        args.extend([no_quote(k) for k in keyids])
        self._handle_io(args, data, result, binary=True)
        logger.debug('send_keys result: %r', result.__dict__)
        data.close()
        return result

    def delete_keys(self, fingerprints, secret=False, passphrase=None,
                    expect_passphrase=True):
        """
        Delete the indicated keys.

        Since GnuPG 2.1, you can't delete secret keys without providing a
        passphrase. However, if you're expecting the passphrase to go to gpg
        via pinentry, you should specify expect_passphrase=False. (It's only
        checked for GnuPG >= 2.1).
        """
        if passphrase and not self.is_valid_passphrase(passphrase):
            raise ValueError('Invalid passphrase')
        which='key'
        if secret:  # pragma: no cover
            if (self.version >= (2, 1) and passphrase is None and
                expect_passphrase):
                raise ValueError('For GnuPG >= 2.1, deleting secret keys '
                                 'needs a passphrase to be provided')
            which='secret-key'
        if _is_sequence(fingerprints):  # pragma: no cover
            fingerprints = [no_quote(s) for s in fingerprints]
        else:
            fingerprints = [no_quote(fingerprints)]
        args = ['--delete-%s' % which]
        if secret and self.version >= (2, 1):
            args.insert(0, '--yes')
        args.extend(fingerprints)
        result = self.result_map['delete'](self)
        if not secret or self.version < (2, 1):
            p = self._open_subprocess(args)
            self._collect_output(p, result, stdin=p.stdin)
        else:
            # Need to send in a passphrase.
            f = _make_binary_stream('', self.encoding)
            try:
                self._handle_io(args, f, result, passphrase=passphrase,
                                binary=True)
            finally:
                f.close()
        return result

    def export_keys(self, keyids, secret=False, armor=True, minimal=False,
                    passphrase=None, expect_passphrase=True):
        """
        Export the indicated keys. A 'keyid' is anything gpg accepts.

        Since GnuPG 2.1, you can't export secret keys without providing a
        passphrase. However, if you're expecting the passphrase to go to gpg
        via pinentry, you should specify expect_passphrase=False. (It's only
        checked for GnuPG >= 2.1).
        """
        if passphrase and not self.is_valid_passphrase(passphrase):
            raise ValueError('Invalid passphrase')
        which=''
        if secret:
            which='-secret-key'
            if (self.version >= (2, 1) and passphrase is None and
                expect_passphrase):
                raise ValueError('For GnuPG >= 2.1, exporting secret keys '
                                 'needs a passphrase to be provided')
        if _is_sequence(keyids):
            keyids = [no_quote(k) for k in keyids]
        else:
            keyids = [no_quote(keyids)]
        args = ['--export%s' % which]
        if armor:
            args.append('--armor')
        if minimal:  # pragma: no cover
            args.extend(['--export-options','export-minimal'])
        args.extend(keyids)
        # gpg --export produces no status-fd output; stdout will be
        # empty in case of failure
        #stdout, stderr = p.communicate()
        result = self.result_map['export'](self)
        if not secret or self.version < (2, 1):
            p = self._open_subprocess(args)
            self._collect_output(p, result, stdin=p.stdin)
        else:
            # Need to send in a passphrase.
            f = _make_binary_stream('', self.encoding)
            try:
                self._handle_io(args, f, result, passphrase=passphrase,
                                binary=True)
            finally:
                f.close()
        logger.debug('export_keys result: %r', result.data)
        # Issue #49: Return bytes if armor not specified, else text
        result = result.data
        if armor:
            result = result.decode(self.encoding, self.decode_errors)
        return result

    def _get_list_output(self, p, kind):
        # Get the response information
        result = self.result_map[kind](self)
        self._collect_output(p, result, stdin=p.stdin)
        lines = result.data.decode(self.encoding,
                                   self.decode_errors).splitlines()
        valid_keywords = 'pub uid sec fpr sub ssb sig'.split()
        for line in lines:
            if self.verbose:  # pragma: no cover
                print(line)
            logger.debug("line: %r", line.rstrip())
            if not line:  # pragma: no cover
                break
            L = line.strip().split(':')
            if not L:  # pragma: no cover
                continue
            keyword = L[0]
            if keyword in valid_keywords:
                getattr(result, keyword)(L)
        return result

    def list_keys(self, secret=False, keys=None, sigs=False):
        """ list the keys currently in the keyring

        >>> import shutil
        >>> shutil.rmtree("keys", ignore_errors=True)
        >>> GPGBINARY = os.environ.get('GPGBINARY', 'gpg')
        >>> if not os.path.isdir('keys'): os.mkdir('keys')
        >>> gpg = GPG(gpgbinary=GPGBINARY, gnupghome='keys')
        >>> input = gpg.gen_key_input(passphrase='foo')
        >>> result = gpg.gen_key(input)
        >>> fp1 = result.fingerprint
        >>> result = gpg.gen_key(input)
        >>> fp2 = result.fingerprint
        >>> pubkeys = gpg.list_keys()
        >>> assert fp1 in pubkeys.fingerprints
        >>> assert fp2 in pubkeys.fingerprints

        """

        if sigs:
            which = 'sigs'
        else:
            which = 'keys'
        if secret:
            which='secret-keys'
        args = ['--list-%s' % which,
                '--fingerprint', '--fingerprint'] # get subkey FPs, too
        if keys:
            if isinstance(keys, string_types):
                keys = [keys]
            args.extend(keys)
        p = self._open_subprocess(args)
        return self._get_list_output(p, 'list')

    def scan_keys(self, filename):
        """
        List details of an ascii armored or binary key file
        without first importing it to the local keyring.

        The function achieves this on modern GnuPG by running:

        $ gpg --dry-run --import-options import-show --import

        On older versions, it does the *much* riskier:

        $ gpg --with-fingerprint --with-colons filename
        """
        if self.version >= (2, 1):
            args = ['--dry-run', '--import-options', 'import-show', '--import']
        else:
            logger.warning('Trying to list packets, but if the file is not a '
                           'keyring, might accidentally decrypt')
            args = ['--with-fingerprint', '--with-colons', '--fixed-list-mode']
        args.append(no_quote(filename))
        p = self._open_subprocess(args)
        return self._get_list_output(p, 'scan')

    def search_keys(self, query, keyserver='pgp.mit.edu'):
        """ search keyserver by query (using --search-keys option)

        >>> import shutil
        >>> shutil.rmtree('keys', ignore_errors=True)
        >>> GPGBINARY = os.environ.get('GPGBINARY', 'gpg')
        >>> if not os.path.isdir('keys'): os.mkdir('keys')
        >>> gpg = GPG(gpgbinary=GPGBINARY, gnupghome='keys')
        >>> os.chmod('keys', 0x1C0)
        >>> result = gpg.search_keys('<vinay_sajip@hotmail.com>')
        >>> if 'NO_EXTERNAL_TESTS' not in os.environ: assert result, 'Failed using default keyserver'
        >>> #keyserver = 'keyserver.ubuntu.com'
        >>> #result = gpg.search_keys('<vinay_sajip@hotmail.com>', keyserver)
        >>> #assert result, 'Failed using keyserver.ubuntu.com'

        """
        query = query.strip()
        if HEX_DIGITS_RE.match(query):
            query = '0x' + query
        args = ['--fingerprint',
                '--keyserver', no_quote(keyserver), '--search-keys',
                no_quote(query)]
        p = self._open_subprocess(args)

        # Get the response information
        result = self.result_map['search'](self)
        self._collect_output(p, result, stdin=p.stdin)
        lines = result.data.decode(self.encoding,
                                   self.decode_errors).splitlines()
        valid_keywords = ['pub', 'uid']
        for line in lines:
            if self.verbose:  # pragma: no cover
                print(line)
            logger.debug('line: %r', line.rstrip())
            if not line:    # sometimes get blank lines on Windows
                continue
            L = line.strip().split(':')
            if not L:  # pragma: no cover
                continue
            keyword = L[0]
            if keyword in valid_keywords:
                getattr(result, keyword)(L)
        return result

    def gen_key(self, input):
        """Generate a key; you might use gen_key_input() to create the
        control input.

        >>> GPGBINARY = os.environ.get('GPGBINARY', 'gpg')
        >>> if not os.path.isdir('keys'): os.mkdir('keys')
        >>> gpg = GPG(gpgbinary=GPGBINARY, gnupghome='keys')
        >>> input = gpg.gen_key_input(passphrase='foo')
        >>> result = gpg.gen_key(input)
        >>> assert result
        >>> result = gpg.gen_key('foo')
        >>> assert not result

        """
        args = ["--gen-key"]
        result = self.result_map['generate'](self)
        f = _make_binary_stream(input, self.encoding)
        self._handle_io(args, f, result, binary=True)
        f.close()
        return result

    def gen_key_input(self, **kwargs):
        """
        Generate --gen-key input per gpg doc/DETAILS
        """
        parms = {}
        for key, val in list(kwargs.items()):
            key = key.replace('_','-').title()
            if str(val).strip():    # skip empty strings
                parms[key] = val
        parms.setdefault('Key-Type','RSA')
        if 'key_curve' not in kwargs:
            parms.setdefault('Key-Length',2048)
        parms.setdefault('Name-Real', "Autogenerated Key")
        logname = (os.environ.get('LOGNAME') or os.environ.get('USERNAME') or
                   'unspecified')
        hostname = socket.gethostname()
        parms.setdefault('Name-Email', "%s@%s" % (logname.replace(' ', '_'),
                                                  hostname))
        out = "Key-Type: %s\n" % parms.pop('Key-Type')
        for key, val in list(parms.items()):
            out += "%s: %s\n" % (key, val)
        out += "%commit\n"
        return out

        # Key-Type: RSA
        # Key-Length: 1024
        # Name-Real: ISdlink Server on %s
        # Name-Comment: Created by %s
        # Name-Email: isdlink@%s
        # Expire-Date: 0
        # %commit
        #
        #
        # Key-Type: DSA
        # Key-Length: 1024
        # Subkey-Type: ELG-E
        # Subkey-Length: 1024
        # Name-Real: Joe Tester
        # Name-Comment: with stupid passphrase
        # Name-Email: joe@foo.bar
        # Expire-Date: 0
        # Passphrase: abc
        # %pubring foo.pub
        # %secring foo.sec
        # %commit

    #
    # ENCRYPTION
    #
    def encrypt_file(self, file, recipients, sign=None,
            always_trust=False, passphrase=None,
            armor=True, output=None, symmetric=False, extra_args=None):
        "Encrypt the message read from the file-like object 'file'"
        if passphrase and not self.is_valid_passphrase(passphrase):
            raise ValueError('Invalid passphrase')
        args = ['--encrypt']
        if symmetric:
            # can't be False or None - could be True or a cipher algo value
            # such as AES256
            args = ['--symmetric']
            if symmetric is not True:
                args.extend(['--cipher-algo', no_quote(symmetric)])
            # else use the default, currently CAST5
        else:
            if not recipients:
                raise ValueError('No recipients specified with asymmetric '
                                 'encryption')
            if not _is_sequence(recipients):
                recipients = (recipients,)
            for recipient in recipients:
                args.extend(['--recipient', no_quote(recipient)])
        if armor:   # create ascii-armored output - False for binary output
            args.append('--armor')
        if output:  # write the output to a file with the specified name
            self.set_output_without_confirmation(args, output)
        if sign is True:  # pragma: no cover
            args.append('--sign')
        elif sign:  # pragma: no cover
            args.extend(['--sign', '--default-key', no_quote(sign)])
        if always_trust:  # pragma: no cover
            args.append('--always-trust')
        if extra_args:
            args.extend(extra_args)
        result = self.result_map['crypt'](self)
        self._handle_io(args, file, result, passphrase=passphrase, binary=True)
        logger.debug('encrypt result: %r', result.data)
        return result

    def encrypt(self, data, recipients, **kwargs):
        """Encrypt the message contained in the string 'data'

        >>> import shutil
        >>> if os.path.exists("keys"):
        ...     shutil.rmtree("keys", ignore_errors=True)
        >>> GPGBINARY = os.environ.get('GPGBINARY', 'gpg')
        >>> if not os.path.isdir('keys'): os.mkdir('keys')
        >>> gpg = GPG(gpgbinary=GPGBINARY, gnupghome='keys')
        >>> input = gpg.gen_key_input(name_email='user1@test', passphrase='pp1')
        >>> result = gpg.gen_key(input)
        >>> fp1 = result.fingerprint
        >>> input = gpg.gen_key_input(name_email='user2@test', passphrase='pp2')
        >>> result = gpg.gen_key(input)
        >>> fp2 = result.fingerprint
        >>> result = gpg.encrypt("hello",fp2)
        >>> message = str(result)
        >>> assert message != 'hello'
        >>> result = gpg.decrypt(message, passphrase='pp2')
        >>> assert result
        >>> str(result)
        'hello'
        >>> result = gpg.encrypt("hello again", fp1)
        >>> message = str(result)
        >>> result = gpg.decrypt(message, passphrase='bar')
        >>> result.status in ('decryption failed', 'bad passphrase')
        True
        >>> assert not result
        >>> result = gpg.decrypt(message, passphrase='pp1')
        >>> result.status == 'decryption ok'
        True
        >>> str(result)
        'hello again'
        >>> result = gpg.encrypt("signed hello", fp2, sign=fp1, passphrase='pp1')
        >>> result.status == 'encryption ok'
        True
        >>> message = str(result)
        >>> result = gpg.decrypt(message, passphrase='pp2')
        >>> result.status == 'decryption ok'
        True
        >>> assert result.fingerprint == fp1

        """
        data = _make_binary_stream(data, self.encoding)
        result = self.encrypt_file(data, recipients, **kwargs)
        data.close()
        return result

    def decrypt(self, message, **kwargs):
        data = _make_binary_stream(message, self.encoding)
        result = self.decrypt_file(data, **kwargs)
        data.close()
        return result

    def decrypt_file(self, file, always_trust=False, passphrase=None,
                     output=None, extra_args=None):
        if passphrase and not self.is_valid_passphrase(passphrase):
            raise ValueError('Invalid passphrase')
        args = ["--decrypt"]
        if output:  # write the output to a file with the specified name
            self.set_output_without_confirmation(args, output)
        if always_trust:  # pragma: no cover
            args.append("--always-trust")
        if extra_args:
            args.extend(extra_args)
        result = self.result_map['crypt'](self)
        self._handle_io(args, file, result, passphrase, binary=True)
        logger.debug('decrypt result: %r', result.data)
        return result

    def trust_keys(self, fingerprints, trustlevel):
        levels = Verify.TRUST_LEVELS
        if trustlevel not in levels:
            poss = ', '.join(sorted(levels))
            raise ValueError('Invalid trust level: "%s" (must be one of %s)' %
                             (trustlevel, poss))
        trustlevel = levels[trustlevel] + 2
        import tempfile
        try:
            fd, fn = tempfile.mkstemp()
            lines = []
            if isinstance(fingerprints, string_types):
                fingerprints = [fingerprints]
            for f in fingerprints:
                lines.append('%s:%s:' % (f, trustlevel))
            # The trailing newline is required!
            s = os.linesep.join(lines) + os.linesep
            logger.debug('writing ownertrust info: %s', s);
            os.write(fd, s.encode(self.encoding))
            os.close(fd)
            result = self.result_map['trust'](self)
            p = self._open_subprocess(['--import-ownertrust', fn])
            self._collect_output(p, result, stdin=p.stdin)
            if p.returncode != 0:
                raise ValueError('gpg returned an error - return code %d' %
                                 p.returncode)
        finally:
            os.remove(fn)
        return result



"""
MAC Address
===========
"""
import re

from insights.core import Parser
from insights.core.exceptions import SkipComponent
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.mac_addresses)
class MacAddress(Parser):
    """
    Parse the machine info in the file "insights-commands/duplidate_machine_info".

    Sample input::

        00:5b:ea:09:00:00

    Attributes:
        address(str): the MAC address

    Examples:
        >>> type(mac_addr)
        <class 'insights.parsers.mac.MacAddress'>
        >>> mac_addr.address
        '00:5b:ea:09:00:00'

    Raises:
        SkipComponent: when there is no expected content in the file
    """
    def parse_content(self, content):
        if not content or len(content) != 1:
            raise SkipComponent('Invalid Content')

        addr = content[0].strip()
        match = re.match("^([0-9a-f]{2}:){5}[0-9a-f]{2}$", addr)
        if match is None:
            raise SkipComponent('Invalid Content')

        self.address = addr



from insights.parsers.password import PasswordAuthPam
from insights.parsers.pam import PamConfEntry
from insights.tests import context_wrap

PW_AUTH_PAM_CONF = """
#%PAM-1.0
# This file is auto-generated.
# User changes will be destroyed the next time authconfig is run.
auth        required      pam_env.so
auth        sufficient    pam_unix.so nullok try_first_pass
auth        requisite     pam_succeed_if.so uid >= 500 quiet
auth        required      pam_deny.so
auth        [default=die] pam_faillock.so authfail deny=3 unlock_time=604800 fail_interval=900
auth        required      pam_faillock.so authsucc deny=3 unlock_time=604800 fail_interval=900

account     required      pam_unix.so
account     sufficient    pam_localuser.so
account     sufficient    pam_succeed_if.so uid < 500 quiet
account     required      pam_permit.so

password    requisite     pam_cracklib.so try_first_pass retry=3 lcredit=-1 ucredit=-1 dcredit=-1 ocredit=-1 minlen=12
password    sufficient    pam_unix.so sha512 shadow nullok try_first_pass use_authtok
password    required      pam_deny.so

session     optional      pam_keyinit.so revoke
session     required      pam_limits.so
session     [success=1 default=ignore] pam_succeed_if.so service in crond quiet use_uid
session     required      pam_unix.so
""".strip()


def test_password_auth_pam_conf():
    pam_conf = PasswordAuthPam(context_wrap(PW_AUTH_PAM_CONF, path='etc/pam.d/password-auth'))
    assert len(pam_conf) == 17
    assert pam_conf[0].service == 'password-auth'
    assert pam_conf[0].interface == 'auth'
    assert len(pam_conf[0].control_flags) == 1
    assert pam_conf[0].control_flags[0].flag == 'required'
    assert pam_conf[0].module_name == 'pam_env.so'
    assert pam_conf[0].module_args is None

    # session     [success=1 default=ignore] pam_succeed_if.so service in crond quiet use_uid
    assert pam_conf[15].interface == 'session'
    assert len(pam_conf[15].control_flags) == 2
    assert pam_conf[15].control_flags == [PamConfEntry.ControlFlag(flag='success', value='1'),
                                         PamConfEntry.ControlFlag(flag='default', value='ignore')]
    assert pam_conf[15].module_name == 'pam_succeed_if.so'
    assert pam_conf[15].module_args == 'service in crond quiet use_uid'



"""
PcpOpenmetricsLog - file ``/var/log/pcp/pmcd/openmetrics.log``
--------------------------------------------------------------
"""
from insights.specs import Specs
from insights import LogFileOutput, parser


@parser(Specs.pcp_openmetrics_log)
class PcpOpenmetricsLog(LogFileOutput):
    """
    Parse the ``/var/log/pcp/pmcd/openmetrics.log`` log file.

    .. note::
        Please refer to its super-class :class:`insights.core.LogFileOutput`

    Sample input::

        [Mon Jul 11 07:59:32] pmdaopenmetrics(1733) Error: cannot parse/store 55% {:
        [Mon Jul 11 07:59:32] pmdaopenmetrics(1733) Error: cannot parse/store }:
        [Mon Jul 11 07:59:32] pmdaopenmetrics(1733) Error: cannot parse/store 95% {:
        [Mon Jul 11 07:59:32] pmdaopenmetrics(1733) Error: cannot parse/store }:
        [Mon Jul 11 07:59:32] pmdaopenmetrics(1733) Error: cannot parse/store };:

    Examples:

        >>> "Error: cannot parse/store" in log
        True
        >>> len(log.get('pmdaopenmetrics')) == 5
        True
    """
    pass



"""
Allow users to interrogate components.

``insights-info foo bar baz`` will search for all datasources that might handle
foo, bar, or baz files or commands along with all components that could be
activated if they were present and valid.

``insights-info -i insights.specs.Specs.hosts`` will display dependency
information about the hosts datasource.

``insights-info -d insights.parsers.hosts.Hosts`` will display the pydoc
information about the Hosts parser.

There are several other options to the script. ``insights-info -h`` for more
info.

"""
from __future__ import print_function
import argparse
import importlib
import inspect
import logging
import pydoc
import re
import sys
import yaml

from six.moves import StringIO

from insights import (apply_default_enabled, apply_configs, dr, get_filters,
        load_default_plugins, load_packages, parse_plugins)
from insights.core.plugins import datasource, is_type
from insights.core import taglang
from insights.core import spec_factory as sf

logging.basicConfig(level=logging.ERROR)


def glob2re(pat):
    """
    Translate a shell PATTERN to a regular expression.
    There is no way to quote meta-characters.

    Stolen from https://stackoverflow.com/a/29820981/1451664
    """

    i, n = 0, len(pat)
    res = ''
    while i < n:
        c = pat[i]
        i = i + 1
        if c == '*':
            # res = res + '.*'
            res = res + '[^/]*'
        elif c == '?':
            # res = res + '.'
            res = res + '[^/]'
        elif c == '[':
            j = i
            if j < n and pat[j] == '!':
                j = j + 1
            if j < n and pat[j] == ']':
                j = j + 1
            while j < n and pat[j] != ']':
                j = j + 1
            if j >= n:
                res = res + '\\['
            else:
                stuff = pat[i:j].replace('\\', '\\\\')
                i = j + 1
                if stuff[0] == '!':
                    stuff = '^' + stuff[1:]
                elif stuff[0] == '^':
                    stuff = '\\' + stuff
                res = '%s[%s]' % (res, stuff)
        else:
            res = res + re.escape(c)
    return res + '\Z(?ms)'


def parse_args():
    p = argparse.ArgumentParser(__doc__.strip())
    p.add_argument("paths", nargs="*", help="Files or commands to use for component activation.")
    p.add_argument("-c", "--components", help="Comma separated list of components that have already executed. Names without '.' are assumed to be in insights.specs.Specs.")
    p.add_argument("--config", help="Configure components.")
    p.add_argument("-i", "--info", help="Comma separated list of components to get dependency info about.", action="store_true")
    p.add_argument("-p", "--preload", help="Comma separated list of packages or modules to preload.")
    p.add_argument("-d", "--pydoc", help="Show pydoc for the given object. E.g.: insights-info -d insights.rule")
    p.add_argument("-k", "--pkg-query", help="Expression to select rules by package.")
    p.add_argument("-s", "--source", help="Show source for the given object. E.g.: insights-info -s insights.core.plugins.rule")
    p.add_argument("-S", "--specs", help="Show specs for the given name.  E.g.: insights-info -S uname", action="store_true")
    p.add_argument("--tags", help="Expression to select rules by tag.")
    p.add_argument("-t", "--types", help="Filter results based on component type; e.g. 'rule,parser'. Names without '.' are assumed to be in insights.core.plugins.")
    p.add_argument("-v", "--verbose", help="Print component dependencies.", action="store_true")
    return p.parse_args()


def load_default_components():
    load_default_plugins()
    dr.load_components("insights.parsers", continue_on_error=False)
    dr.load_components("insights.combiners", continue_on_error=False)


def preload_components(comps):
    plugins = parse_plugins(comps)
    load_packages(plugins)


def get_components(comps, default_module):
    if not default_module.endswith("."):
        default_module = default_module + "."

    results = []
    if comps:
        comps = [t.strip() for t in comps.split(",")]
        comps = [t if "." in t else default_module + t for t in comps]
        for t in comps:
            c = dr.get_component(t)
            if not c:
                raise Exception("Could not load %s." % t)
            results.append(c)
    return results


def get_datasources():
    ds_types = (sf.simple_file,
                sf.glob_file,
                sf.simple_command,
                sf.first_file,
                sf.foreach_execute,
                sf.foreach_collect)
    ds = dr.COMPONENTS_BY_TYPE[datasource]
    return [d for d in ds if isinstance(d, ds_types)]


def matches(d, path):
    path = path.strip('/')

    if isinstance(d, sf.simple_file):
        # Set the filename from the simple_file string as search target
        search_tgt = d.path.strip('/').split()[0].split('/')[-1]
        return search_tgt.startswith(path)

    if isinstance(d, sf.glob_file):
        for pat in d.patterns:
            # Set the last part of the glob_file as the search target
            pat_sp = pat.split('/')
            search_tgt = pat_sp[-1]
            if search_tgt.startswith("*") and len(pat_sp) >= 2:
                # if the last part is "*" or "*XXX", set the parent dir as search target
                # e.g: xinetd_conf: "/etc/xinet.d/*.conf"
                #      logrotate_conf: "/etc/logrotate.d/*"
                search_tgt = pat_sp[-2]
                return search_tgt.startswith(path) and not d.ignore_func(path)
            # Glob check the last part: `startswith` or glob match
            if ((search_tgt.startswith(path) or
                 re.match(glob2re(search_tgt), path)) and
                    not d.ignore_func(path)):
                return True
        return False

    if isinstance(d, sf.simple_command):
        # Get the command name from the simple_command string
        search_tgt = d.cmd.strip('/').split()[0].split('/')[-1]
        return search_tgt.startswith(path)

    if isinstance(d, sf.first_file):
        for p in d.paths:
            # Get the filename as the search target
            search_tgt = p.strip('/').split()[0].split('/')[-1]
            if search_tgt.startswith(path):
                return True
        return False

    if isinstance(d, sf.foreach_execute):
        # Get the command name from the simple_command string
        search_tgt = d.cmd.strip('/').split()[0].split('/')[-1] if d.cmd else ''
        return search_tgt.startswith(path)

    if isinstance(d, sf.foreach_collect):
        # Get the filename from the "glob_file" like string
        search_tgt = d.path.strip('/').split()[0].split('/')[-1]
        return search_tgt.startswith(path) and not d.ignore_func(path)


def get_matching_datasources(paths):
    if not paths:
        return []
    ds = get_datasources()
    results = set()
    for path in paths:
        for d in ds:
            if matches(d, path):
                results.add(d)
    return list(results)


def create_broker(components):
    broker = dr.Broker()
    for c in components:
        broker[c] = 1
    return broker


def dry_run(graph=dr.COMPONENTS[dr.GROUPS.single], broker=None):
    broker = broker or dr.Broker()
    for c in broker.instances:
        yield c
    for c in dr.run_order(graph):
        d = dr.get_delegate(c)
        if d and d.get_missing_dependencies(broker) is None:
            broker[c] = 1
            yield c


def dump_ds(d, space=""):
    dbl_space = space * 2
    delegate = dr.get_delegate(d)
    try:
        print(space + "Class: %s" % d.__class__)
    except:
        pass

    try:
        print(space + "Filtered: %s" % delegate.filterable)
    except:
        pass

    print(space + "Raw: %s" % delegate.raw)
    print(space + "Multioutput: %s" % delegate.multi_output)

    if isinstance(d, sf.simple_file):
        print(space + "Path: %s" % d.path)

    if isinstance(d, (sf.simple_command, sf.foreach_execute)):
        print(space + "Command: %s" % d.cmd)

    if isinstance(d, sf.first_file):
        print(space + "Paths:")
        for p in d.paths:
            print(dbl_space + p)

    if isinstance(d, sf.glob_file):
        print(space + "Patterns:")
        for p in d.patterns:
            print(dbl_space + p)
        print(space + "Ignore: %s" % d.ignore)

    if isinstance(d, sf.foreach_collect):
        print(space + "Path: %s" % d.path)
        print(space + "Ignore: %s" % d.ignore)

    filters = get_filters(d)
    if filters:
        print(space + "Filters:")
        for f in filters:
            print(dbl_space + f)


def print_component(comp, verbose=False, specs=False):
    if (specs or (not specs and not is_type(comp, datasource))):
        print(dr.get_name(comp))

    if not verbose:
        return

    space = " " * 4
    dbl_space = space * 2

    d = dr.get_delegate(comp)
    print(space + "Type: %s" % dr.get_name(d.type))
    if is_type(comp, datasource):
        dump_ds(comp, space=space)

    print()
    if d.requires:
        print(space + "Requires:")
        for r in d.requires:
            print(dbl_space + dr.get_name(r))
    else:
        print(space + "Requires: nothing")

    if d.at_least_one and d.at_least_one[0]:
        for one in d.at_least_one:
            print(space + "At least one of:")
            for o in one:
                print(dbl_space + dr.get_name(o))

    if d.optional:
        print(space + "Optional:")
        for r in d.optional:
            print(dbl_space + dr.get_name(r))

    dependents = dr.get_dependents(comp)
    if dependents:
        print(space + "Dependents:")
        for r in sorted(dependents, key=dr.get_name):
            print(dbl_space + dr.get_name(r))
    tags = dr.get_tags(comp)
    if tags:
        print(space + "Tags:")
        for r in sorted(tags):
            print(dbl_space + r)
    print()


def print_results(results, types, verbose, specs):
    for r in results:
        if not types or is_type(r, types):
            print_component(r, verbose=verbose, specs=specs)


def dump_info(comps):
    for c in comps:
        comp = dr.get_component(c)
        if not comp:
            print("Unknown component: %s" % c)
        else:
            print_component(comp, verbose=True, specs=True)


def load_obj(spec):
    try:
        return dr.get_component(spec) or importlib.import_module(spec)
    except:
        pass


def get_source(spec):
    obj = load_obj(spec)
    if obj:
        try:
            return inspect.getsource(obj)
        except:
            pass


def get_pydoc(spec):
    obj = load_obj(spec)
    if obj:
        output = StringIO()
        pydoc.Helper(output=output).help(obj)
        output.seek(0)
        return output.read()


def apply_configuration(path):
    with open(path) as f:
        config = yaml.safe_load(f)
        load_packages(config.get('packages', []))
        apply_default_enabled(config)
        apply_configs(config)


def main():
    if "" not in sys.path:
        sys.path.insert(0, "")

    args = parse_args()

    if args.source:
        src = get_source(args.source)
        print(src or "{} has no python source.".format(args.source))
        return

    if args.pydoc:
        doc = get_pydoc(args.pydoc)
        print(doc or "{} has no pydoc.".format(args.pydoc))
        return

    if args.info:
        if args.config:
            apply_configuration(args.config)
        dump_info(args.paths)
        return

    load_default_components()

    if args.preload:
        preload_components(args.preload)

    if args.config:
        apply_configuration(args.config)

    types = tuple(get_components(args.types, "insights.core.plugins"))

    if args.pkg_query:
        pred = taglang.parse(args.pkg_query)

        def pkg_query(c):
            return pred([dr.get_module_name(c)])
    else:
        def pkg_query(c):
            return True

    if args.tags:
        pred = taglang.parse(args.tags)

        def tags_query(c):
            return pred(dr.get_tags(c))
    else:
        def tags_query(c):
            return True

    components = get_components(args.components, "insights.specs.Specs")

    if not args.paths:
        components = [c for c in list(dr.DELEGATES) if pkg_query(c) and tags_query(c)]
        print_results(components, types, args.verbose, args.specs)
        return

    ds = get_matching_datasources(args.paths)

    broker = create_broker(components + ds)
    results = [r for r in dry_run(broker=broker) if pkg_query(r) and tags_query(r)]
    print_results(results, types, args.verbose, args.specs)


if __name__ == "__main__":
    main()



"""
Block device listing
====================

Module for processing output of the ``lsblk`` command.  Different information
is provided by the ``lsblk`` command depending upon the options.  Parsers
included here are:

LSBlock - Command ``lsblk``
---------------------------

The ``LSBlock`` class parses output of the ``lsblk`` command with no options.

LSBlockPairs - Command ``lsblk -P -o [columns...]``
---------------------------------------------------

The ``LSBlockPairs`` class parses output of the ``lsblk -P -o [columns...]``
command.

These classes based on ``BlockDevices`` which implements all of the
functionality except the parsing of command specific information.
Information is stored in the attribute ``self.rows`` which is a ``list`` of
``BlockDevice`` objects.

Each ``BlockDevice`` object provides the functionality for one row of data from the
command output.  Data in a ``BlockDevice`` object is accessible by multiple methods.
For example the NAME field can be accessed in the following four ways::

    lsblk_info.rows[0].data['NAME']
    lsblk_info.rows[0].NAME
    lsblk_info.rows[0].name
    lsblk_info.rows[0].get('NAME')

Sample output of the ``lsblk`` command looks like::

    NAME          MAJ:MIN RM  SIZE RO TYPE MOUNTPOINT
    vda           252:0    0    9G  0 disk
    |-vda1        252:1    0  500M  0 part /boot
    `-vda2        252:2    0  8.5G  0 part
      |-rhel-root 253:0    0  7.6G  0 lvm  /
      |-rhel-swap 253:1    0  924M  0 lvm  [SWAP]
    sda             8:0    0  500G  0 disk
    `-sda1          8:1    0  500G  0 part /data

Note the hierarchy demonstrated in the name column. For instance ``vda1`` and
``vda2`` are children of ``vda``.  Likewise, ``rhel-root`` and ``rhel-swap``
are children of ``vda2``.  This relationship is demonstrated in the
``PARENT_NAMES`` key, which is only present if the row is a *child* row.  For
example ``PARENT_NAMES`` value for ``rhel-root`` will be ``['vda', 'vda2']``
meaning that ``vda2`` is the immediate parent and ``vda`` is parent of
``vda2``.

Also note that column names that are not valid Python property names been
changed.  For example ``MAJ:MIN`` has been changed to ``MAJ_MIN``.

Examples:
    >>> lsblk_info = shared[LSBlock]
    >>> lsblk_info
    <insights.parsers.lsblk.LSBlock object at 0x7f1f6a422d50>
    >>> lsblk_info.rows
    [disk:vda,
     part:vda1(/boot),
     part:vda2,
     lvm:rhel-root(/),
     lvm:rhel-swap([SWAP]),
     disk:sda,
     part:sda1(/data)]
    >>> lsblk_info.rows[0]
    disk:vda
    >>> lsblk_info.rows[0].data
    {'READ_ONLY': False, 'NAME': 'vda', 'REMOVABLE': False, 'MAJ_MIN': '252:0',
     'TYPE': 'disk', 'SIZE': '9G'}
    >>> lsblk_info.rows[0].data['NAME']
    'vda'
    >>> lsblk_info.rows[0].NAME
    'vda'
    >>> lsblk_info.rows[0].name
    'vda'
    >>> lsblk_info.rows[0].data['MAJ_MIN']
    '252:0'
    >>> lsblk_info.rows[0].MAJ_MIN
    '252:0'
    >>> lsblk_info.rows[0].maj_min
    '252:0'
    >>> lsblk_info.rows[0].removable
    False
    >>> lsblk_info.rows[0].read_only
    False
    >>> lsblk_info.rows[2].data
    {'READ_ONLY': False, 'PARENT_NAMES': ['vda'], 'NAME': 'vda2',
     'REMOVABLE': False, 'MAJ_MIN': '252:2', 'TYPE': 'part', 'SIZE': '8.5G'}
    >>> lsblk_info.rows[2].parent_names
    ['vda']
    >>> lsblk_info.rows[3].parent_names
    ['vda', 'vda2']
    >>> lsblk_info.device_data['vda'] # Access devices by name
    'disk:vda'
    >>> lsblk_info.search(NAME='vda2')
    [{'READ_ONLY': False, 'PARENT_NAMES': ['vda'], 'NAME': 'vda2',
     'REMOVABLE': False, 'MAJ_MIN': '252:2', 'TYPE': 'part', 'SIZE': '8.5G'}]
"""
from __future__ import division
import re

from insights.core import CommandParser
from insights.core.exceptions import ParseException
from insights.core.plugins import parser
from insights.parsers import keyword_search
from insights.specs import Specs

MAX_GENERATIONS = 20


class BlockDevice(object):
    """Class to contain one line of ``lsblk`` command information.

    Contains all of the fields for a single line of ``lsblk`` output.
    Computed values are the column names except where the column
    name is an invalid variable name in Python such as `MAJ:MIN`.
    The ``get`` method is provided to access any value, including
    those that are not valid names in Python.  All other valid
    names may be accessed as ``obj.column_name``.
    """
    def __init__(self, data):
        self.data = data
        for k, v in data.items():
            k = re.sub(r'[-:\.]', "_", k)
            setattr(self, k, v)
            setattr(self, k.lower(), v)

    def __contains__(self, item):
        return hasattr(self, item)

    def __eq__(self, other):
        return self.data == other

    def iteritems(self):
        return self.items()

    def items(self):
        return self.data.items()

    def get(self, k, default=None):
        """Get any value by keyword (column) name."""
        return self.data.get(k, default)

    def __str__(self):
        if 'TYPE' in self.data and 'MOUNTPOINT' in self.data:
            return '{type}:{name}({mnt})'.format(
                type=self.data['TYPE'], name=self.data['NAME'],
                mnt=self.data['MOUNTPOINT']
            )
        else:
            # As long as the regular expression in LsBlock works, we must end
            # up with NAME and TYPE records here.  In LSBlockPairs this is
            # enforced with an explicit check.
            return '{type}:{name}'.format(type=self.data['TYPE'], name=self.data['NAME'])


class BlockDevices(CommandParser):
    """Class to contain all information from ``lsblk`` command.

    Output of the ``lsblk`` command is contained in this base
    class. Data may be accessed via the iterator and each item
    represents a row of output from the command in `dict` format.

    Attributes:
        rows (list of BlockDevice): List of ``BlockDevice`` objects for each
            row of the input. Input column name matches key name except any
            '-' is replaced with '_' and the following names are changed::

                Column Name     Key Name
                MAJ:MIN         MAJ_MIN
                RM              REMOVABLE
                RO              READD_ONLY
        device_data (dict of BlockDevice): A dictionary of ``BlockDevice``
            objects keyed on the 'NAME' column (e.g. ``sda`` or ``rhel-swap``)
    """

    def __len__(self):
        return len(self.rows)

    def __iter__(self):
        for row in self.rows:
            yield row

    def search(self, **kwargs):
        """
        Returns a list of the block devices (in order) matching the given
        criteria. Keys are searched for directly - see the
        :py:func:`insights.parsers.keyword_search` utility function for more
        details.  If no search parameters are given, no rows are returned.
        Keys need to be in all upper case, as they appear in the source data.

        Examples:

            >>> blockdevs.search(NAME='sda1')
            [{'NAME': '/dev/sda1', 'TYPE': 'disk', 'SIZE', '80G', ...}]
            >>> blockdevs.search(TYPE='lvm')
            [{'NAME': 'volgrp01-root', 'TYPE': 'lvm', 'SIZE', '15G', ...}...]

        Arguments:
            **kwargs (dict): Dictionary of key-value pairs to search for.

        Returns:
            (list): The list of mount points matching the given criteria.
        """
        return keyword_search(self.rows, **kwargs)


@parser(Specs.lsblk)
class LSBlock(BlockDevices):
    """Parse output of the ``lsblk`` command.

    The specific lsblk commands are ``/bin/lsblk`` and ``/usr/bin/lsblk``.
    Typical content of the ``lsblk`` command output looks like::

        NAME                            MAJ:MIN RM  SIZE RO TYPE  MOUNTPOINT
        sda                               8:0    0   80G  0 disk
        |-sda1                            8:1    0  256M  0 part  /boot
        `-sda2                            8:2    0 79.8G  0 part
          |-volgrp01-root (dm-0)        253:0    0   15G  0 lvm   /
          `-volgrp01-swap (dm-1)        253:1    0    8G  0 lvm   [SWAP]

    Note:
        See the discussion of the key ``PARENT_NAMES`` above.
    """
    def parse_content(self, content):
        r = re.compile(r"([\s\|\`\-]*)(\S+.*) (\d+:\d+)\s+(\d)\s+(\d+(\.\d)?[A-Z])\s+(\d)\s+([a-z0-9]+)(.*)")
        device_list = []
        parents = [None] * MAX_GENERATIONS
        for line in content[1:]:
            name_match = r.match(line)
            generation = 0
            if name_match and len(name_match.groups()) == 9:
                device = {}
                name = name_match.group(2).strip()
                generation = len(name_match.group(1)) // 2
                parents[generation] = name
                device['NAME'] = name
                device['MAJ_MIN'] = name_match.group(3)
                device['REMOVABLE'] = bool(int(name_match.group(4)))
                device['SIZE'] = name_match.group(5)
                device['READ_ONLY'] = bool(int(name_match.group(7)))
                # TYPE is enforced by the regex, no need to check here
                device['TYPE'] = name_match.group(8)
                mountpoint = name_match.group(9).strip()
                if len(mountpoint) > 0:
                    device['MOUNTPOINT'] = mountpoint
                if generation > 0:
                    device['PARENT_NAMES'] = parents[:generation]
                device_list.append(device)

        self.rows = [BlockDevice(d) for d in device_list]
        self.device_data = dict((dev.name, dev) for dev in self.rows)


@parser(Specs.lsblk_pairs)
class LSBlockPairs(BlockDevices):
    """Parse output of the ``lsblk -P -o`` command.

    ``lsblk`` command with ``-P -o`` options provides explicit selection of
    output columns in keyword=value pairs.

    The specific lsblk commands are ``/bin/lsblk -P -o column_names`` and
    ``/usr/bin/lsblk -P -o column_names``.  Typical content of the ``lsblk``
    command output looks like::

        ALIGNMENT="0" DISC-ALN="0" DISC-GRAN="0B" DISC-MAX="0B" DISC-ZERO="0" \
            FSTYPE="" GROUP="cdrom" KNAME="sr0" LABEL="" LOG-SEC="512" MAJ:MIN="11:0" \
            MIN-IO="512" MODE="brw-rw----" MODEL="DVD+-RW DVD8801 " MOUNTPOINT="" \
            NAME="sr0" OPT-IO="0" OWNER="root" PHY-SEC="512" RA="128" RM="1" RO="0" \
            ROTA="1" RQ-SIZE="128" SCHED="cfq" SIZE="1024M" STATE="running" TYPE="rom" UUID=""
        ALIGNMENT="0" DISC-ALN="0" DISC-GRAN="0B" DISC-MAX="0B" DISC-ZERO="0" \
            FSTYPE="" GROUP="disk" KNAME="sda" LABEL="" LOG-SEC="512" MAJ:MIN="8:0" \
            MIN-IO="512" MODE="brw-rw----" MODEL="WDC WD1600JS-75N" MOUNTPOINT="" \
            NAME="sda" OPT-IO="0" OWNER="root" PHY-SEC="512" RA="128" RM="0" RO="0" \
            ROTA="1" RQ-SIZE="128" SCHED="cfq" SIZE="149G" STATE="running" TYPE="disk" UUID=""
        ALIGNMENT="0" DISC-ALN="0" DISC-GRAN="0B" DISC-MAX="0B" DISC-ZERO="0" \
            FSTYPE="ext4" GROUP="disk" KNAME="sda1" LABEL="" LOG-SEC="512" MAJ:MIN="8:1" \
            MIN-IO="512" MODE="brw-rw----" MODEL="" MOUNTPOINT="/boot" NAME="sda1" \
            OPT-IO="0" OWNER="root" PHY-SEC="512" RA="128" RM="0" RO="0" ROTA="1" \
            RQ-SIZE="128" SCHED="cfq" SIZE="500M" STATE="" TYPE="part" \
            UUID="c7c4c016-8b00-4ded-bffb-5cc4719b7d45"

    Attributes:
        rows (list of BlockDevice): List of ``BlockDevice`` objects for each row of
            the input. Input column name matches key name except that
            any '-', ':', or '.' is replaced with
            '_' and the following names are changed::

                Column Name     Key Name
                RM              removable
                RO              read_only
        failed_device_paths (set): Set of device names that failed to get device path

    Note:
        ``PARENT_NAMES`` is not available as a key because it is not listed
        in the ``LsBlockPairs`` output and cannot always be correctly
        inferred from the other data present.
    """
    def parse_content(self, content):
        self.rows = []
        self.failed_device_paths = set()
        if "invalid option" in content[0] and "lsblk:" in content[0]:
            raise ParseException(content[0])
        for line in content:
            if ' TYPE=' not in line:
                if 'failed to get device path' in line:
                    self.failed_device_paths.add(line.split(":")[1].strip())
                    continue
                else:
                    raise ParseException(
                        "TYPE not found in LsBlockPairs line '{l}'".format(l=line)
                    )

            d = dict((k, v) for k, v in re.findall(r'(\S+)=\"(.*?)\"\s?', line) if len(v) > 0)

            def str2bool(s):
                return bool(int(s))

            for original, replace, transform in [("RM", "REMOVABLE", str2bool),
                                                 ("RO", "READ_ONLY", str2bool)]:
                if original in d:
                    d[replace] = transform(d[original]) if transform else d[original]
                    del d[original]
            self.rows.append(BlockDevice(d))
        self.device_data = dict((dev.name, dev) for dev in self.rows)



"""
LsVarTmp - command ``ls -ln /var/tmp``
======================================

The ``ls -ln /var/tmp`` command provides information for the listing of the
``/var/tmp`` directory.

Sample input is shown in the Examples. See ``FileListing`` class for
additional information.

Sample directory list::

    /var/tmp:
    total 20
    drwxr-xr-x.  2 0 0 4096 Mar 26 02:25 a1
    drwxr-xr-x.  2 0 0 4096 Mar 26 02:25 a2
    drwxr-xr-x.  2 0 0 4096 Apr 28  2018 foreman-ssh-cmd-fc3f65c9-2b35-480d-87e3-1d971433d6ad

Examples:

    >>> "a1" in ls_var_tmp
    False
    >>> "/var/tmp" in ls_var_tmp
    True
    >>> ls_var_tmp.dir_entry('/var/tmp', 'a1')['type']
    'd'
"""


from insights.specs import Specs
from insights.core.filters import add_filter
from insights.util import deprecated

from .. import FileListing
from .. import parser, CommandParser


add_filter(Specs.ls_var_tmp, "/var/tmp")


@parser(Specs.ls_var_tmp)
class LsVarTmp(CommandParser, FileListing):
    """
    .. warning::
        This class is deprecated and will be removed from 3.5.0.
        Please use the :class:`insights.parsers.ls.LSlanFiltered` instead.

    Parses output of ``ls -ln /var/tmp`` command.
    """
    def __init__(self, *args, **kwargs):
        deprecated(LsVarTmp, "Please use the :class:`insights.parsers.ls.LSlanFiltered` instead.", "3.5.0")
        super(LsVarTmp, self).__init__(*args, **kwargs)



"""
NormalMD5 Combiner for the NormalMD5 Parser
===========================================

Combiner for the :class:`insights.parsers.md5check.NormalMD5` parser.

This parser is multioutput, one parser instance for each file md5sum.
Ths combiner puts all of them back together and presents them as a dict
where the keys are the filenames and the md5sums are the values.

This class inherits all methods and attributes from the ``dict`` object.

Examples:
    >>> type(md5sums)
    <class 'insights.combiners.md5check.NormalMD5'>
    >>> sorted(md5sums.keys())
    ['/etc/localtime1', '/etc/localtime2']
    >>> md5sums['/etc/localtime2']
    'd41d8cd98f00b204e9800998ecf8427e'
"""

from .. import combiner
from insights.parsers.md5check import NormalMD5 as NormalMD5Parser


@combiner(NormalMD5Parser)
class NormalMD5(dict):
    """
    Combiner for the NormalMD5 parser.
    """
    def __init__(self, md5_checksums):
        super(NormalMD5, self).__init__()
        for md5info in md5_checksums:
            self.update({md5info.filename: md5info.md5sum})



import doctest
import pytest

from insights.core.exceptions import SkipComponent
from insights.parsers import ovs_vsctl
from insights.parsers.ovs_vsctl import OVSvsctlListBridge
from insights.tests import context_wrap

OVS_VSCTL_LIST_BRIDGES_ALL = """
_uuid               : xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
auto_attach         : []
controller          : [xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx]
datapath_id         : "0000a61fd19ea54f"
datapath_type       : "enp0s9"
datapath_version    : "<unknown>"
external_ids        : {a="0"}
fail_mode           : secure
flood_vlans         : [1000]
flow_tables         : {1=xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx}
ipfix               : xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
mcast_snooping_enable: false
mirrors             : [xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx]
name                : br-int
netflow             : xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
other_config        : {disable-in-band="true", mac-table-size="2048"}
ports               : [xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx, 0000000-0000-0000-0000-0000000000000, 1111111-1111-1111-1111-1111111111111]
protocols           : ["OpenFlow11", "OpenFlow11", "OpenFlow12", "OpenFlow13"]
rstp_enable         : true
rstp_status         : {rstp_bridge_id="8.000.a61fd19ea54f",     rstp_bridge_port_id="0000", rstp_designated_id="8.000.a61fd19ea54f", rstp_designated_port_id="0000", rstp_root_id="8.000.a61fd19ea54f", rstp_root_path_cost="0"}
sflow               : xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
status              : {"0"="1"}
stp_enable          : true

_uuid               : xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
auto_attach         : []
controller          : []
datapath_id         : "0000d29e6a8acc4c"
datapath_type       : ""
datapath_version    : "<unknown>"
external_ids        : {}
fail_mode           : []
flood_vlans         : []
flow_tables         : {}
ipfix               : []
mcast_snooping_enable: false
mirrors             : []
name                : br-tun
netflow             : []
other_config        : {}
ports               : [xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx]
protocols           : []
rstp_enable         : false
rstp_status         : {}
sflow               : []
status              : {}
stp_enable          : false
""".strip()

OVS_VSCTL_LIST_BRIDGES_FILTERED1 = """
name                : br-int
other_config        : {disable-in-band="true", mac-table-size="2048"}
name                : br-tun
other_config        : {}
""".strip()

OVS_VSCTL_LIST_BRIDGES_FILTERED2 = """
_uuid               : xxxxxxxx-xxxx-xxxx-xxxxx-xxxxxxxxxxxx
name                : br-int
netflow             : []
other_config        : {disable-in-band="true", mac-table-size="2048"}
stp_enable          : false
_uuid               : aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa
name                : br-tun
netflow             : []
other_config        : {mac-table-size="4096"}
stp_enable          : true
""".strip()

EXCEPTION1 = """
""".strip()


def test_ovs_vsctl_documentation():
    env = {
        "bridge_lists": OVSvsctlListBridge(context_wrap(OVS_VSCTL_LIST_BRIDGES_FILTERED1)),
    }
    failed, total = doctest.testmod(ovs_vsctl, globs=env)
    assert failed == 0


def test_ovs_vsctl_all():
    bridges = OVSvsctlListBridge(context_wrap(OVS_VSCTL_LIST_BRIDGES_ALL))
    assert bridges[0]["name"] == "br-int"
    assert bridges[0]["external_ids"] == {"a": "0"}
    assert bridges[0]["flow_tables"] == {"1": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx"}
    assert bridges[0].get("flood_vlans") == ["1000"]
    assert bridges[0]["protocols"][-1] == "OpenFlow13"
    assert bridges[0]["other_config"]["mac-table-size"] == "2048"
    assert bridges[0]["rstp_status"]["rstp_root_path_cost"] == "0"
    assert bridges[1]["name"] == "br-tun"
    assert bridges[1]["mirrors"] == []
    assert bridges[1]["datapath_type"] == ""
    assert bridges[1]["status"] == {}
    assert bridges[1].get("stp_enable") == "false"


def test_ovs_vsctl():
    bridges = OVSvsctlListBridge(context_wrap(OVS_VSCTL_LIST_BRIDGES_FILTERED2))
    assert bridges[0].get("name") == "br-int"
    assert bridges[0]["other_config"]["mac-table-size"] == "2048"
    assert bridges[1]["name"] == "br-tun"


def test_ovs_vsctl_exception1():
    with pytest.raises(SkipComponent) as e:
        OVSvsctlListBridge(context_wrap(EXCEPTION1))
    assert "Empty file" in str(e)



import doctest
import pytest

from insights.core.exceptions import SkipComponent
from insights.parsers import ld_library_path
from insights.parsers.ld_library_path import UserLdLibraryPath
from insights.tests import context_wrap

LD_LIBRARY_PATH_EMPTY = """
""".strip()

LD_LIBRARY_PATH_INVALID = """
LD_LIBRARY_PATH: Undefined variable.
""".strip()

LD_LIBRARY_PATH_DOC = """
sr1adm /usr/sap/RH1/SYS/exe/run:/usr/sap/RH1/SYS/exe/uc/linuxx86_64:/sapdb/clients/RH1/lib
sr2adm
rh1adm /usr/sap/RH1/SYS/exe/uc/linuxx86_64:/usr/sap/RH1/SYS/exe/run
""".strip()

LD_LIBRARY_PATH = """
sr1adm /usr/sap/RH1/SYS/exe/run:/usr/sap/RH1/SYS/exe/uc/linuxx86_64:/sapdb/clients/RH1/lib
sr2adm "/usr/sap/RH1/SYS/exe/uc/linuxx86_64:/usr/sap/RH1/SYS/exe/run"
sr3adm
rh1adm ''
""".strip()  # noqa: W391


def test_ld_library_path():
    ret = UserLdLibraryPath(context_wrap(LD_LIBRARY_PATH))
    assert len(ret) == 4
    assert ret[0].raw == LD_LIBRARY_PATH.splitlines()[0].split()[-1]
    assert ret[1].raw == LD_LIBRARY_PATH.splitlines()[1].split()[-1]
    assert ret[2].user == 'sr3adm'
    assert ret[2].raw == ''
    assert ret[2].path == ['']
    assert ret[3].user == 'rh1adm'
    assert ret[3].raw == "''"
    assert ret[3].path == ['']
    for p in LD_LIBRARY_PATH.splitlines()[0].split()[-1].split(':'):
        assert p in ret[0].path
    for p in LD_LIBRARY_PATH.splitlines()[1].split()[-1].strip('"').split(':'):
        assert p in ret[1].path


def test_empty_and_invalid():
    with pytest.raises(SkipComponent):
        UserLdLibraryPath(context_wrap(LD_LIBRARY_PATH_EMPTY))


def test_doc_examples():
    env = {
        'ld_lib_path': UserLdLibraryPath(context_wrap(LD_LIBRARY_PATH_DOC)),
    }
    failed, total = doctest.testmod(ld_library_path, globs=env)
    assert failed == 0



# -*- coding: utf-8 -*-

from threading import Timer
from multiprocessing import Process
from os import getenv
from requests import Session
from requests.adapters import BaseAdapter
from requests.exceptions import ConnectionError, ConnectTimeout
from time import sleep, time


class Timeout:
    """
    Runs a function for a limited amount of time. Reports whether it finished in time.
    """
    PRISTINE = 0
    RUNNING = 1
    SUCCESS = 2
    FAIL = 3
    TIMEOUT = 4

    def __init__(self):
        """
        New timeout service. self.state becomes PRISTINE.
        """
        self.reset()

    def reset(self):
        """
        Prepares the service for the next run, setting the state to PRISTINE.
        """
        self.state = self.PRISTINE

    def start(self, worker, timeout):
        """
        Spawns given function in a subprocess. Regularly checks whether it has finished or whether the set time-out has
        been reached. Sets self.state accordingly. Terminates the process on time-out.
        """
        def tick():
            """
            Relaunches itself every second until the process finishes or time runs out.
            """
            if not process.is_alive():
                self.state = self.FAIL if process.exitcode else self.SUCCESS
                return

            now = time()
            lapsed = now - started
            if lapsed > timeout:
                process.terminate()
                self.state = self.TIMEOUT
                return

            timer = Timer(1, tick)
            timer.start()
            timer.join()

        self.reset()

        process = Process(target=worker)
        process.start()
        started = time()
        self.state = self.RUNNING

        tick()


class DummyRequestsAdapter(BaseAdapter):
    """
    Emulates timeout behavior without making actual requests. Supports only simple numeric timeout values.
    """
    def send(self, request, **kwargs):
        if kwargs["timeout"] is not None:
            sleep(kwargs["timeout"])
            raise ConnectTimeout
        else:
            sleep(60)  # One minute, don't hang for eternity.
            raise ConnectionError

    def close(self):
        pass


def dummy_requests_session():
    """
    Mount our dummy adapter to every requests.Session instance created.
    """
    instance = Session()
    adapter = DummyRequestsAdapter()
    instance.mount("https://", adapter)  # All our requests are HTTPS.
    return instance


def getenv_bool(name, default=None):
    """
    Get a boolean value from an environment variable. Accepts only True or False.
    """
    env_string = getenv(name)
    if env_string == "True":
        bool_value = True
    elif env_string == "False":
        bool_value = False
    elif env_string is None:
        if default is None:
            raise ValueError("{0} must be set. Provide True/False.".format(name))
        bool_value = default
    else:
        raise ValueError("{0}={1} is not a valid value. Provide True/False.".format(name, env_string))
    return bool_value



"""
YumLog - file ``/var/log/yum.log``
==================================

This module provides parsing for the ``/var/log/yum.log`` file.
"""
from collections import defaultdict, namedtuple
from insights.core import Parser
from insights.core.exceptions import ParseException, SkipComponent
from insights.core.filters import add_filter
from insights.core.plugins import parser
from insights.parsers.installed_rpms import InstalledRpm
from insights.specs import Specs


Entry = namedtuple('Entry', field_names='idx timestamp state pkg')
"""namedtuple: Represents a package line in ``/var/log/yum.log``."""
VALID_STATES = ["Erased:", "Installed:", "Updated:"]
"""Package states in this file"""
add_filter(Specs.yum_log, VALID_STATES)


@parser(Specs.yum_log)
class YumLog(Parser, list):
    """
    Class for parsing file ``/var/log/yum.log`` into a list of .

    The file looks like::

        May 13 15:54:49 Installed: libevent-2.0.21-4.el7.x86_64
        May 13 15:54:49 Installed: tmux-1.8-4.el7.x86_64
        May 23 18:06:24 Installed: wget-1.14-10.el7_0.1.x86_64
        May 23 18:10:05 Updated: 1:openssl-libs-1.0.1e-51.el7_2.5.x86_64
        May 23 18:10:05 Installed: 1:perl-parent-0.225-244.el7.noarch
        May 23 18:10:05 Installed: perl-HTTP-Tiny-0.033-3.el7.noarch
        May 23 16:09:09 Erased: redhat-access-insights-batch
        May 23 18:10:05 Installed: perl-podlators-2.5.1-3.el7.noarch
        May 23 18:10:05 Installed: perl-Pod-Perldoc-3.20-4.el7.noarch
        May 23 18:10:05 Installed: 1:perl-Pod-Escapes-1.04-286.el7.noarch
        May 23 18:10:06 Installed: perl-Text-ParseWords-3.29-4.el7.noarch

    The information is stored as a ``list`` of :py:data:`Entry` objects, each
    of which contains attributes for the position in the log, timestamp of the
    action, the package's state in the system, and the affected package as an
    :py:class:`insights.parsers.installed_rpms.InstalledRpm`.

    .. note::
        The file will be filtered at least with the keywords defined in the
        :py:data:`STATES`, that means only the lines that contain one of
        these keywords will be collected.

    Examples:
        >>> type(yl)
        <class 'insights.parsers.yumlog.YumLog'>
        >>> e = yl.present_packages.get('nss-softokn')
        >>> e.pkg.release == '19.el6_6'
        True
        >>> e = yl.present_packages.get('openssl-libs')
        >>> e.pkg.name == 'openssl-libs'
        True
        >>> e.pkg.version == '1.0.1e'
        True
        >>> len(yl)
        8
        >>> sorted(e.idx for e in yl) == sorted(range(len(yl)))
        True
    """
    ERASED = 'Erased'
    """
    .. deprecated:: 3.2.0

       Will be removed from 3.4.0.  Use string "Erased" instead.
    """

    INSTALLED = 'Installed'
    """
    .. deprecated:: 3.2.0

       Will be removed from 3.4.0.  Use string "Installed" instead.
    """

    UPDATED = 'Updated'
    """
    .. deprecated:: 3.2.0

       Will be removed from 3.4.0.  Use string "Updated" instead.
    """

    STATES = set(["Erased", "Installed", "Updated"])

    def parse_content(self, content):
        """Parses contents of each line in ``/var/log/yum.log``.

        Each line in the file contains 5 fields that are parsed into the
        attributes of :py:data:`Entry` instances:

            - month
            - day
            - time
            - state
            - package

        The month, day, and time form the ``Entry.timestamp``.
        ``Entry.state`` contains the state of the package, one of ``Erased``,
        ``Installed``, or ``Updated``. ``Entry.pkg`` contains the
        :py:class:`insights.parsers.installed_rpms.InstalledRpm` instance
        corresponding to the parse package.

        ``Entry.idx`` is the zero-based line number (but not the RAW line number)
        of the ``Entry`` in the file.  It can ONLY be used to tell ordering
        of events.

        Parameters:
            content (list): Lines of ``/var/log/yum.log`` to be parsed.

        Raises:
            ParseException: if a line can't be parsed for any reason.
            SkipComponent: if nothing is parsed.
        """
        self._pkgs = defaultdict(list)
        for idx, line in enumerate(content):
            if not any(s in line for s in self.STATES):  # Do not remove
                continue
            try:
                line = line.replace(': 100', '')
                month, day, time, state, pkg = line.split()[:5]
                timestamp = ' '.join([month, day, time])
                state = state.rstrip(':')
                pkg = pkg.split(':')[-1].strip()
                if "." not in pkg:
                    pkg = InstalledRpm({'name': pkg})
                else:
                    pkg = InstalledRpm.from_package(pkg)
                e = Entry(idx, timestamp, state, pkg)
                self.append(e)
                self._pkgs[pkg.name].append(e)
            except:
                raise ParseException('YumLog could not parse', line)

        if len(self) == 0:
            raise SkipComponent

        self.__cur_pkgs = dict((k, v[-1]) for k, v in self._pkgs.items())

    @property
    def pkgs(self):
        """
        .. deprecated:: 3.2.0

           Will be removed from 3.4.0.
        """
        return dict(self._pkgs)

    @property
    def data(self):
        """
        .. deprecated:: 3.2.0

           Will be removed from 3.4.0.
        """
        return self

    @property
    def present_packages(self):
        """`list` of latest ``Entry`` instances for installed packages."""
        return dict((k, v) for k, v in self.__cur_pkgs.items() if v.state in ['Installed', 'Updated'])

    def packages_of(self, state):
        """
        Parameters:
            state: one state or list of states of `Installed`, `Updated`, and `Erased`.

        Returns:
            `list` of latest ``Entry`` instances for packages in `state`.

        Raises:
            KeyError: when getting invalid state.
        """
        state = [state] if isinstance(state, str) else state
        if any(st not in self.STATES for st in state):
            raise KeyError('Invalid State(s): {0}'.format(state))
        return dict((k, v) for k, v in self.__cur_pkgs.items() if v.state in state)



"""
LsBoot - command ``ls -lanR /boot``
===================================

The ``ls -lanR /boot`` command provides information for the listing of the
``/boot`` directory.

See :class:`insights.parsers.ls.FileListing` for more information.

"""

from insights import CommandParser, parser
from insights.parsers.ls import FileListing
from insights.specs import Specs


@parser(Specs.ls_boot)
class LsBoot(CommandParser, FileListing):
    """
    Parse the /boot directory listing using a standard FileListing parser.

    .. warning::

        For Insights Advisor Rules, it's recommended to use the
        :class:`insights.parsers.ls.LSlanR` and add the ``"/boot"`` to
        the filter list of `Specs.ls_lanR_dirs` instead.

    Sample directory listing::

        /boot:
        total 187380
        dr-xr-xr-x.  3 0 0     4096 Mar  4 16:19 .
        dr-xr-xr-x. 19 0 0     4096 Jul 14 09:10 ..
        -rw-r--r--.  1 0 0   123891 Aug 25  2015 config-3.10.0-229.14.1.el7.x86_64

        /boot/grub2:
        total 36
        drwxr-xr-x. 6 0 0  104 Mar  4 16:16 .
        dr-xr-xr-x. 3 0 0 4096 Mar  4 16:19 ..
        lrwxrwxrwx. 1 0 0     11 Aug  4  2014 menu.lst -> ./grub.conf
        -rw-r--r--. 1 0 0   64 Sep 18  2015 device.map

    Examples:
        >>> type(bootdir)
        <class 'insights.parsers.ls_boot.LsBoot'>
        >>> '/boot' in bootdir
        True
        >>> '/boot/grub' in bootdir
        False
        >>> bootdir.files_of('/boot')
        ['config-3.10.0-229.14.1.el7.x86_64']
        >>> bootdir.dirs_of('/boot')
        ['.', '..', 'grub2']
        >>> bootdir.dir_contains('/boot/grub2', 'menu.lst')
        True
    """
    __root_path = '/boot'



from insights.tests import context_wrap
from insights.parsers.sysconfig import ChronydSysconfig


CHRONYD = """
OPTIONS="-d"
#HIDE="me"
""".strip()


def test_sysconfig_chronyd():
    result = ChronydSysconfig(context_wrap(CHRONYD))
    assert result['OPTIONS'] == '-d'
    assert result.get('HIDE') is None



"""
Cron jobs - files ``/etc/cron.d/*``
===================================

This module contains the following parsers:

CronForeman - file ``/etc/cron.d/foreman``
------------------------------------------
"""
from insights.core import Parser
from insights.core.exceptions import ParseException
from insights.core.plugins import parser
from insights.parsers import get_active_lines, keyword_search
from insights.specs import Specs


class CronFile(Parser):
    """
    It parses the file at ``/etc/cron.d/*```.

    Each row of the file is converted into a dictionary with keys for each field.
    And also the whole line is saved for use to the "raw" key.
    For example one row would look like::

        {
          'minute': '*',
          'hour': '*',
          'day_of_month': '*',
          'month': '*',
          'day_of_week': '*',
          'username': 'test',
          'command': '/usr/bin/keystone-manage token_flush > /dev/null 2>&1',
          'raw': '* * * * * * test /usr/bin/keystone-manage token_flush > /dev/null 2>&1'
        }

    It parses the line in the same way that cron(1) does.  Lines that
    are blank or start with a comment are ignored.  Environment lines of the
    form 'KEY = value' (with optional spacing around the equals sign) are
    stored in the `environment` dictionary. Lines containing a valid
    cron job, with five recogniseable time fields, username, a command,
    are stored in the `jobs` property and accessed through the pseudo-list
    interface and the `search` method. All other lines are stored in the
    `invalid_lines` property.

    It also recognises the extension of time signature 'nicknames', which
    take place of the first five parts of a standard crontab line:

        * `@reboot` : Run once after reboot.
        * `@yearly` : Run once a year, ie.  "0 0 1 1 \*".
        * `@annually` : Run once a year, ie.  "0 0 1 1 \*".
        * `@monthly` : Run once a month, ie. "0 0 1 \* \*".
        * `@weekly` : Run once a week, ie.  "0 0 \* \* 0".
        * `@daily` : Run once a day, ie.   "0 0 \* \* \*".
        * `@hourly` : Run once an hour, ie. "0 \* \* \* \*".

    In the case of the '@reboot' nickname, the row will not contain the 'minute',
    'hour', 'day_of_month', 'month', or 'day_of_week' keys, and instead will
    contain the key 'time' with the value '@reboot' (as well as the usual
    'username' and 'command' key). All other nicknames are translated directly
    into their five-part equivalent and parsed as a normal crontab line.

    Sample input::

        SHELL=/bin/sh
        TEST_HOME=/usr/share/test

        15 23 * * *     test    /usr/sbin/testaa 2>&1 >> /var/log/test.log
        0 7 * * *       test    /usr/sbin/testbb 2>&1 >> /var/log/test.log

    Attributes:
        jobs(list): A list of valid jobs info.
        environment(dict): The environment variables.
        invalid_lines(list): A list of lines which are neither in cron format nor in env variable format.

    Raises:
        ParseException: When a line starts with '@', but the nickname is unknown.

    Examples:
        >>> type(cron_obj)
        <class 'insights.parsers.cron_jobs.CronFile'>
        >>> len(cron_obj.environment)
        2
        >>> 'TEST_HOME' in cron_obj.environment
        True
        >>> cron_obj.environment['TEST_HOME']
        '/usr/share/test'
        >>> len(cron_obj.jobs)
        2
        >>> cron_obj.jobs[0].get('minute')
        '15'
        >>> cron_obj.jobs[0].get('hour')
        '23'
        >>> cron_obj.jobs[0].get('username')
        'test'
        >>> '/usr/sbin/testaa' in cron_obj.jobs[0].get('command')
        True
    """
    def parse_content(self, content):
        self.jobs = []
        self.environment = {}
        self.invalid_lines = []
        # 'nicknames' can be used for common event frequencies:
        nicknames = {
            '@yearly': '0 0 1 1 *',
            '@annually': '0 0 1 1 *',
            '@monthly': '0 0 1 * *',
            '@weekly': '0 0 * * 0',
            '@daily': '0 0 * * *',
            '@hourly': '0 * * * *',
        }
        for line in get_active_lines(content):
            if line.startswith('@'):
                # Reboot is 'special':
                if line.startswith('@reboot'):
                    parts = line.split(None, 2)
                    self.jobs.append({'time': '@reboot', 'username': parts[1], 'command': parts[2], 'raw': line})
                    continue
                else:
                    parts = line.split(None, 2)
                    if parts[0] not in nicknames:
                        raise ParseException(
                            "{n} not recognised as a time specification 'nickname'".format(n=parts[0])
                        )
                    # Otherwise, put the time spec nickname translation into
                    # the line
                    line = line.replace(parts[0], nicknames[parts[0]])
            items = line.split(None, 6)
            if len(items) == 7:
                self.jobs.append({
                    'minute': items[0],
                    'hour': items[1],
                    'day_of_month': items[2],
                    'month': items[3],
                    'day_of_week': items[4],
                    'username': items[5],
                    'command': items[6],
                    'raw': line
                })
            else:
                parts = line.split('=')
                if len(parts) == 2:
                    self.environment[parts[0].strip()] = parts[1].strip()
                else:
                    self.invalid_lines.append(line)

    def search(self, **kwargs):
        """
        Return a list of dict for lines that matches the kwargs.

        This uses the :py:func:`insights.parsers.keyword_search` function for
        searching; see its documentation for usage details. If no search
        parameters are given, a empty list is returned.
        """
        return keyword_search(self.jobs, **kwargs)


@parser(Specs.cron_foreman)
class CronForeman(CronFile):
    """
    Parse the ``/etc/cron.d/foreman`` file.

    Sample input::

        SHELL=/bin/sh
        PATH=/usr/local/sbin:/usr/local/bin:/sbin:/bin:/usr/sbin:/usr/bin

        RAILS_ENV=production
        FOREMAN_HOME=/usr/share/foreman

        # Clean up the session entries in the database
        15 23 * * *     foreman    /usr/sbin/foreman-rake db:sessions:clear 2>&1 | gawk '{ print strftime("[\%Y-\%m-\%d \%H:\%M:\%S]"), $0 }' >>/var/log/foreman/cron.log

        # Send out recurring notifications
        0 7 * * *       foreman    /usr/sbin/foreman-rake reports:daily 2>&1 | gawk '{ print strftime("[\%Y-\%m-\%d \%H:\%M:\%S]"), $0 }' >>/var/log/foreman/cron.log
        0 5 * * 0       foreman    /usr/sbin/foreman-rake reports:weekly 2>&1 | gawk '{ print strftime("[\%Y-\%m-\%d \%H:\%M:\%S]"), $0 }' >>/var/log/foreman/cron.log

    Examples:
        >>> type(foreman_jobs)
        <class 'insights.parsers.cron_jobs.CronForeman'>
        >>> len(foreman_jobs.environment)
        4
        >>> 'FOREMAN_HOME' in foreman_jobs.environment
        True
        >>> foreman_jobs.environment['FOREMAN_HOME']
        '/usr/share/foreman'
        >>> len(foreman_jobs.jobs)
        3
        >>> foreman_jobs.jobs[0].get('minute')
        '15'
        >>> foreman_jobs.jobs[0].get('hour')
        '23'
        >>> foreman_jobs.jobs[0].get('username')
        'foreman'
        >>> 'db:sessions:clear' in foreman_jobs.jobs[0].get('command')
        True
    """
    pass



"""
WatchDogLog - file ``/var/log/watchdog/*.std*``
===============================================
"""

from insights import LogFileOutput, parser
from insights.specs import Specs


@parser(Specs.watchdog_logs)
class WatchDogLog(LogFileOutput):
    """
    Class for parsing ``/var/log/watchdog/*.std*`` files.

    Sample Input::

        DEBUG:root:0

        INFO:root:Executing: /usr/bin/sg_persist -n -i -k -d /dev/mapper/mpathb

        DEBUG:root:0   PR generation=0xc2, 4 registered reservation keys follow:
            0xfdfe0001
            0xfdfe0001
            0xfdfe0000
            0xfdfe0000


        DEBUG:root:key fdfe0001 registered with device /dev/mapper/mpathb
        INFO:root:Executing: /usr/bin/sg_turs /dev/mapper/mpathb

        DEBUG:root:0
    """

    time_format = None



"""
SAPHostCtrlInstances - Commands ``saphostctrl``
===============================================
"""
from insights.core import CommandParser
from insights.core.exceptions import ParseException, SkipComponent
from insights.core.filters import add_filter
from insights.core.plugins import parser
from insights.specs import Specs


SAP_INST_FILTERS = [
        '******',
        'SID',
        'SystemNumber',
        'InstanceName',
        'InstanceType',
        'Hostname',
        'SapVersionInfo',
        'FullQualifiedHostname'
]
add_filter(Specs.saphostctl_getcimobject_sapinstance, SAP_INST_FILTERS)


@parser(Specs.saphostctl_getcimobject_sapinstance)
class SAPHostCtrlInstances(CommandParser, list):
    """
    This class provides processing for the output of the
    ``saphostctrl -function GetCIMObject -enuminstances SAPInstance``
    command on SAP systems.

    Sample output of the command::

        *********************************************************
         SID , String , D89
         SystemNumber , String , 88
         InstanceName , String , HDB88
         InstanceType , String , HANA Test
         Hostname , String , hdb88
         FullQualifiedHostname , String , hdb88.example.com
         IPAddress , String , 10.0.0.88
         SapVersionInfo , String , 749, patch 211, changelist 1754007
        *********************************************************
         SID , String , D90
         SystemNumber , String , 90
         InstanceName , String , HDB90
         InstanceType , String , HANA Test
         Hostname , String , hdb90
         FullQualifiedHostname , String , hdb90.example.com
         IPAddress , String , 10.0.0.90
         SapVersionInfo , String , 749, patch 211, changelist 1754007

    Examples:
        >>> type(sap_inst)
        <class 'insights.parsers.saphostctrl.SAPHostCtrlInstances'>
        >>> sap_inst[1]['SID']
        'D90'
        >>> sap_inst[0]['SapVersionInfo']
        '749, patch 211, changelist 1754007'
        >>> sap_inst[1]['InstanceType']
        'HANA Test'
        >>> sap_inst.types
        ['HDB']
        >>> sorted(sap_inst.sids)
        ['D89', 'D90']

    Raises:
        SkipComponent: When input is empty.
        ParseException: When input cannot be parsed.
    """
    def parse_content(self, content):
        self._instances = []
        self._sids = set()
        self._types = set()

        inst = {}
        for line in content:
            line = line.strip()
            if not line:
                continue
            if line.startswith('******'):
                inst = {}
                self.append(inst)
                continue

            fields = [i.strip() for i in line.split(',', 2)]
            if len(fields) < 3:
                raise ParseException("Incorrect line: '{0}'".format(line))

            inst[fields[0]] = fields[2]

            if fields[0] == 'InstanceName':
                short_type = fields[2].strip('0123456789')
                # Back forward compatible: a short InstanceType was from the InstanceName
                if 'InstanceType' not in inst:
                    inst.update(InstanceType=short_type)
                self._instances.append(fields[2])
                self._types.add(short_type)

            self._sids.add(fields[2]) if fields[0] == 'SID' else None

        if len(self) < 1:
            raise SkipComponent("Nothing need to parse")

    @property
    def instances(self):
        """
        (list): The list of instances found in the cluster output.
        """
        return self._instances

    @property
    def sids(self):
        """
        (list): The list of SID found in the cluster output.
        """
        return sorted(self._sids)

    @property
    def types(self):
        """
        (list): The list of short instance types.  The short instance
                type is set as the characters of the `InstanceName` before
                the `SystemNumber`.
        """
        return sorted(self._types)

    @property
    def data(self):
        """
        (list): List of dicts where keys are the lead name of each line and
                values are the string value.
        """
        return self



import doctest
import pytest

from insights.core.exceptions import SkipComponent
from insights.parsers import sudoers
from insights.parsers.sudoers import EtcSudoers
from insights.tests import context_wrap

SUDOERS = """
## Allows people in group wheel to run all commands
%wheel  ALL=(ALL)       ALL
## Read drop-in files from /etc/sudoers.d (the # here does not mean a comment)
#includedir /etc/sudoers.d
""".strip()
SUDOERS_EMPTY = ""


def test_etc_sudoers():
    sudo = EtcSudoers(context_wrap(SUDOERS))
    assert len(sudo.lines) == 2
    assert sudo.get(['wheel', 'ALL=(ALL)', 'ALL']) == ['%wheel  ALL=(ALL)       ALL']
    assert sudo.last("#includedir") == '#includedir /etc/sudoers.d'
    assert len(sudo.get(['wheel', 'includedir'])) == 0
    assert len(sudo.get(['wheel', 'includedir'], check=any)) == 2
    assert sudo.last(['wheel', 'includedir'], check=any) == '#includedir /etc/sudoers.d'


def test_ab():
    with pytest.raises(SkipComponent):
        EtcSudoers(context_wrap(SUDOERS_EMPTY))

    sudo = EtcSudoers(context_wrap(SUDOERS))
    with pytest.raises(TypeError):
        sudo.get({})


def test_doc_examples():
    env = {
        'sudo': EtcSudoers(context_wrap(SUDOERS))
    }
    failed, total = doctest.testmod(sudoers, globs=env)
    assert failed == 0
